
#include "cmdlib.h"
#include "mathlib.h"
#include "bspfile.h"
#include "mem.h"

extern qboolean noskip;

//=============================================================================

int			nummodels;
dmodel_t	dmodels[MAX_MAP_MODELS];

int			visdatasize;
byte		dvisdata[MAX_MAP_VISIBILITY];

int			lightdatasize;
byte		dlightdata[MAX_MAP_LIGHTING];

// LordHavoc: stored in .lit file
int			rgblightdatasize;
byte		drgblightdata[MAX_MAP_LIGHTING*3];

// LordHavoc: stored in .dlit file
int			nmaplightdatasize;
byte		dnmaplightdata[MAX_MAP_LIGHTING*3];


int			texdatasize;
byte		dtexdata[MAX_MAP_MIPTEX]; // (dmiptexlump_t)

int			entdatasize;
char		dentdata[MAX_MAP_ENTSTRING];

int			numleafs;
dleaf_t		dleafs[MAX_MAP_LEAFS];

int			numplanes;
dplane_t	dplanes[MAX_MAP_PLANES];

int			numvertexes;
dvertex_t	dvertexes[MAX_MAP_VERTS];

int			numnodes;
dnode_t		dnodes[MAX_MAP_NODES];

int			numtexinfo;
texinfo_t	texinfo[MAX_MAP_TEXINFO];

int			numfaces;
dface_t		dfaces[MAX_MAP_FACES];

int			numclipnodes;
dclipnode_t	dclipnodes[MAX_MAP_CLIPNODES];

int			numedges;
dedge_t		dedges[MAX_MAP_EDGES];

int			nummarksurfaces;
unsigned int		dmarksurfaces[MAX_MAP_MARKSURFACES];

int			numsurfedges;
int			dsurfedges[MAX_MAP_SURFEDGES];

hullinfo_t	hullinfo;

//=============================================================================

typedef struct
{
	int		initialized;
	int		maxsize;
	byte	*start;
	byte	*index;
} swappedbuffer_t;

void SB_LoadFile (swappedbuffer_t *sbuf, char *filename)
{
	FILE	*f;

	f = SafeOpenRead (filename);
	sbuf->maxsize = Q_filelength (f);
	sbuf->start = qmalloc (sbuf->maxsize + 1);
	sbuf->start[sbuf->maxsize] = 0;
	SafeRead (f, sbuf->start, sbuf->maxsize);
	fclose (f);

	sbuf->index = sbuf->start;
	sbuf->initialized = true;
}

void SB_Alloc (swappedbuffer_t *sbuf, int maxsize)
{
	sbuf->maxsize = maxsize;
	sbuf->start = qmalloc (maxsize);
	sbuf->index = sbuf->start;
	sbuf->initialized = true;
}

void SB_Free (swappedbuffer_t *sbuf)
{
	if (!sbuf->initialized)
		return;

	qfree (sbuf->start);
	memset (sbuf, 0, sizeof (*sbuf));
}

byte SB_ReadByte (swappedbuffer_t *sbuf)
{
	byte b;
	b = *sbuf->index++;
	return b;
}

short SB_ReadShort (swappedbuffer_t *sbuf)
{
	short s;
	s = sbuf->index[0] + 256 * sbuf->index[1];
	sbuf->index += 2;
	return s;
}

int SB_ReadInt (swappedbuffer_t *sbuf)
{
	int i;
	i = sbuf->index[0] + 256 * sbuf->index[1] + 65536 * sbuf->index[2] + 16777216 * sbuf->index[3];
	sbuf->index += 4;
	return i;
}

float SB_ReadFloat (swappedbuffer_t *sbuf)
{
	union
	{
		int		i;
		float	f;
	} u;

	u.i = SB_ReadInt (sbuf);
	return u.f;
}

void SB_ReadData (swappedbuffer_t *sbuf, void *d, int n)
{
	memcpy (d, sbuf->index, n);
	sbuf->index += n;
}

void SB_ZeroFill (swappedbuffer_t *sbuf, int n)
{
	while (n--)
		*sbuf->index++ = 0;
}

void SB_WriteByte (swappedbuffer_t *sbuf, byte b)
{
	*sbuf->index++ = b;
}

void SB_WriteShort (swappedbuffer_t *sbuf, short i)
{
	sbuf->index[0] = i & 0xFF;
	sbuf->index[1] = (i >> 8) & 0xFF;
	sbuf->index += 2;
}

void SB_WriteInt (swappedbuffer_t *sbuf, int i)
{
	sbuf->index[0] = i & 0xFF;
	sbuf->index[1] = (i >> 8) & 0xFF;
	sbuf->index[2] = (i >> 16) & 0xFF;
	sbuf->index[3] = (i >> 24) & 0xFF;
	sbuf->index += 4;
}

void SB_WriteFloat (swappedbuffer_t *sbuf, float f)
{
	union
	{
		int		i;
		float	f;
	} u;

	u.f = f;
	SB_WriteInt (sbuf, u.i);
}

void SB_WriteData (swappedbuffer_t *sbuf, void *d, int n)
{
	memcpy (sbuf->index, d, n);
	sbuf->index += n;
}

void SB_WriteString (swappedbuffer_t *sbuf, char *s)
{
	while (*s)
		*sbuf->index++ = *s++;

	*sbuf->index++ = 0;
}

void SB_SeekAbsolute (swappedbuffer_t *sbuf, int index)
{
	sbuf->index = sbuf->start + index;
}

int SB_Tell (swappedbuffer_t *sbuf)
{
	return sbuf->index - sbuf->start;
}


void SwapBSPFile (qboolean todisk)
{
	int		i, j, c;
	dmodel_t	*d;
	dmiptexlump_t	*mtl;

	// models
	for (i = 0; i < nummodels; i++)
	{
		d = &dmodels[i];

		for (j = 0; j < MAX_MAP_HULLS; j++)
			d->headnode[j] = LittleLong (d->headnode[j]);

		d->visleafs = LittleLong (d->visleafs);
		d->firstface = LittleLong (d->firstface);
		d->numfaces = LittleLong (d->numfaces);

		for (j = 0; j < 3; j++)
		{
			d->mins[j] = LittleFloat (d->mins[j]);
			d->maxs[j] = LittleFloat (d->maxs[j]);
			d->origin[j] = LittleFloat (d->origin[j]);
		}
	}

	// vertexes
	for (i = 0; i < numvertexes; i++)
	{
		for (j = 0; j < 3; j++)
			dvertexes[i].point[j] = LittleFloat (dvertexes[i].point[j]);
	}

	// planes
	for (i = 0; i < numplanes; i++)
	{
		for (j = 0; j < 3; j++)
			dplanes[i].normal[j] = LittleFloat (dplanes[i].normal[j]);

		dplanes[i].dist = LittleFloat (dplanes[i].dist);
		dplanes[i].type = LittleLong (dplanes[i].type);
	}

	// texinfos
	for (i = 0; i < numtexinfo; i++)
	{
		for (j = 0; j < 8; j++)
			texinfo[i].vecs[0][j] = LittleFloat (texinfo[i].vecs[0][j]);

		texinfo[i].miptex = LittleLong (texinfo[i].miptex);
		texinfo[i].flags = LittleLong (texinfo[i].flags);
	}

	// faces
	for (i = 0; i < numfaces; i++)
	{
		dfaces[i].texinfo = LittleLong (dfaces[i].texinfo);
		dfaces[i].planenum = LittleLong (dfaces[i].planenum);
		dfaces[i].side = LittleShort (dfaces[i].side);
		dfaces[i].lightofs = LittleLong (dfaces[i].lightofs);
		dfaces[i].firstedge = LittleLong (dfaces[i].firstedge);
		dfaces[i].numedges = LittleLong (dfaces[i].numedges);
	}

	// nodes
	for (i = 0; i < numnodes; i++)
	{
		dnodes[i].planenum = LittleLong (dnodes[i].planenum);

		for (j = 0; j < 3; j++)
		{
			dnodes[i].mins[j] = LittleShort (dnodes[i].mins[j]);
			dnodes[i].maxs[j] = LittleShort (dnodes[i].maxs[j]);
		}

		dnodes[i].children[0] = LittleLong (dnodes[i].children[0]);
		dnodes[i].children[1] = LittleLong (dnodes[i].children[1]);
		dnodes[i].firstface = LittleLong (dnodes[i].firstface);
		dnodes[i].numfaces = LittleLong (dnodes[i].numfaces);
	}

	// leafs
	for (i = 0; i < numleafs; i++)
	{
		dleafs[i].contents = LittleLong (dleafs[i].contents);

		for (j = 0; j < 3; j++)
		{
			dleafs[i].mins[j] = LittleShort (dleafs[i].mins[j]);
			dleafs[i].maxs[j] = LittleShort (dleafs[i].maxs[j]);
		}

		dleafs[i].firstmarksurface = LittleLong (dleafs[i].firstmarksurface);
		dleafs[i].nummarksurfaces = LittleLong (dleafs[i].nummarksurfaces);
		dleafs[i].visofs = LittleLong (dleafs[i].visofs);
	}

	// clipnodes
	for (i = 0; i < numclipnodes; i++)
	{
		dclipnodes[i].planenum = LittleLong (dclipnodes[i].planenum);
		dclipnodes[i].children[0] = LittleLong (dclipnodes[i].children[0]);
		dclipnodes[i].children[1] = LittleLong (dclipnodes[i].children[1]);
	}

	// miptex
	if (texdatasize)
	{
		mtl = (dmiptexlump_t *) dtexdata;

		if (todisk)
			c = mtl->nummiptex;
		else c = LittleLong (mtl->nummiptex);

		mtl->nummiptex = LittleLong (mtl->nummiptex);

		for (i = 0; i < c; i++)
			mtl->dataofs[i] = LittleLong (mtl->dataofs[i]);
	}

	// marksurfaces
	for (i = 0; i < nummarksurfaces; i++)
		dmarksurfaces[i] = LittleLong (dmarksurfaces[i]);

	// surfedges
	for (i = 0; i < numsurfedges; i++)
		dsurfedges[i] = LittleLong (dsurfedges[i]);

	// edges
	for (i = 0; i < numedges; i++)
	{
		dedges[i].v[0] = LittleLong (dedges[i].v[0]);
		dedges[i].v[1] = LittleLong (dedges[i].v[1]);
	}
}


/*
=============
LoadBSPFile
=============
*/
int SB_ReadLump (swappedbuffer_t *sb, lump_t *lump, void *dest, int size)
{
	SB_SeekAbsolute (sb, lump->fileofs);
	SB_ReadData (sb, dest, lump->filelen);

	return lump->filelen / size;
}


void LoadBSPFile (char *filename)
{
	int				i, j;
	swappedbuffer_t	sb;
	lump_t			lumps[HEADER_LUMPS], *lump;

	// load file into buffer
	SB_LoadFile (&sb, filename);

	// hull 0 is always point-sized
	VectorClear (hullinfo.hullsizes[0][0]);
	VectorClear (hullinfo.hullsizes[0][1]);

	// check header
	i = SB_ReadInt (&sb);

	if (i != BSPVERSION)
		Error ("%s is version %i, should be %i", filename, i, BSPVERSION);

	hullinfo.numhulls = 3;
	hullinfo.filehulls = 4;

	VectorSet (hullinfo.hullsizes[1][0], -16, -16, -24);
	VectorSet (hullinfo.hullsizes[1][1], 16, 16, 32);
	VectorSet (hullinfo.hullsizes[2][0], -32, -32, -24);
	VectorSet (hullinfo.hullsizes[2][1], 32, 32, 64);

	for (i = 0; i < HEADER_LUMPS; i++)
	{
		lumps[i].fileofs = SB_ReadInt (&sb);
		lumps[i].filelen = SB_ReadInt (&sb);
	}

	// read lumps (sigh...)
	numplanes = SB_ReadLump (&sb, &lumps[LUMP_PLANES], dplanes, sizeof (dplane_t));
	numleafs = SB_ReadLump (&sb, &lumps[LUMP_LEAFS], dleafs, sizeof (dleaf_t));
	numvertexes = SB_ReadLump (&sb, &lumps[LUMP_VERTEXES], dvertexes, sizeof (dvertex_t));
	numnodes = SB_ReadLump (&sb, &lumps[LUMP_NODES], dnodes, sizeof (dnode_t));
	numtexinfo = SB_ReadLump (&sb, &lumps[LUMP_TEXINFO], texinfo, sizeof (texinfo_t));
	numfaces = SB_ReadLump (&sb, &lumps[LUMP_FACES], dfaces, sizeof (dface_t));
	numclipnodes = SB_ReadLump (&sb, &lumps[LUMP_CLIPNODES], dclipnodes, sizeof (dclipnode_t));
	nummarksurfaces = SB_ReadLump (&sb, &lumps[LUMP_MARKSURFACES], dmarksurfaces, sizeof (dmarksurfaces[0]));
	numsurfedges = SB_ReadLump (&sb, &lumps[LUMP_SURFEDGES], dsurfedges, sizeof (dsurfedges[0]));
	numedges = SB_ReadLump (&sb, &lumps[LUMP_EDGES], dedges, sizeof (dedge_t));
	nummodels = SB_ReadLump (&sb, &lumps[LUMP_MODELS], dmodels, sizeof (dmodel_t));
	lightdatasize = SB_ReadLump (&sb, &lumps[LUMP_LIGHTING], dlightdata, 1);
	visdatasize = SB_ReadLump (&sb, &lumps[LUMP_VISIBILITY], dvisdata, 1);
	entdatasize = SB_ReadLump (&sb, &lumps[LUMP_ENTITIES], dentdata, 1);
	texdatasize = SB_ReadLump (&sb, &lumps[LUMP_TEXTURES], dtexdata, 1);

	// finish up
	SB_Free (&sb);
}


void SB_WriteLump (swappedbuffer_t *sb, lump_t *lump, void *data, int len)
{
	lump->fileofs = SB_Tell (sb);
	SB_WriteData (sb, data, len);
	lump->filelen = SB_Tell (sb) - lump->fileofs;
}


int skipcount;

void SkipBSPFile (void)
{
	int i, j, k;
	unsigned int *leafmarks;
	dleaf_t *leaf;
	dface_t *face;
	texinfo_t *ti;
	miptex_t *textures, *texture;
	char *name;
	dmiptexlump_t *miptexlump;
	dmodel_t *model;
	dface_t *modfaces;

	miptexlump = (dmiptexlump_t *) dtexdata;
	textures = (miptex_t *) dtexdata;
	skipcount = 0;

	for (i = 0, leaf = dleafs; i < numleafs; i++, leaf++)
	{
		leafmarks = dmarksurfaces + leaf->firstmarksurface;

		for (j = 0; j < leaf->nummarksurfaces;)
		{
			face = dfaces + leafmarks[j];
			ti = texinfo + face->texinfo;
			texture = (miptex_t *) ((byte *) textures + miptexlump->dataofs[ti->miptex]);
			name = texture->name;

			if (!strcmp (name, "skip") ||
				!strcmp (name, "*waterskip") ||
				!strcmp (name, "*slimeskip") ||
				!strcmp (name, "*lavaskip"))
			{
				// copy each remaining marksurface to previous slot
				for (k = j; k < leaf->nummarksurfaces - 1; k++)
					leafmarks[k] = leafmarks[k + 1];

				// reduce marksurface count by one
				leaf->nummarksurfaces--;

				skipcount++;
			}
			else j++;
		}
	}

	// loop through all the models, editing surface lists (this takes care of brush entities)
	for (i = 0, model = dmodels; i < nummodels; i++, model++)
	{
		if (i == 0) continue; // model 0 is worldmodel

		modfaces = dfaces + model->firstface;

		for (j = 0; j < model->numfaces;)
		{
			face = modfaces + j;
			ti = texinfo + face->texinfo;
			texture = (miptex_t *) ((byte *) textures + miptexlump->dataofs[ti->miptex]);
			name = texture->name;

			if (!strcmp (name, "skip") ||
				!strcmp (name, "*waterskip") ||
				!strcmp (name, "*slimeskip") ||
				!strcmp (name, "*lavaskip"))
			{
				// copy each remaining face to previous slot
				for (k = j; k < model->numfaces - 1; k++)
					modfaces[k] = modfaces[k + 1];

				// reduce face count by one
				model->numfaces--;

				skipcount++;
			}
			else j++;
		}
	}

	qprintf ("Skipped %i surfaces", skipcount);
}


/*
=============
WriteBSPFile

Swaps the bsp file in place, so it should not be referenced again
=============
*/
void WriteBSPFile (char *filename, qboolean litonly)
{
	int				i, j;
	FILE			*f;
	swappedbuffer_t	sb;

	if (!litonly)
	{
		int		index;
		int		bspsize;
		lump_t	lumps[HEADER_LUMPS], *lump;

		if (!noskip) SkipBSPFile ();

		// bleagh - the original hard-coded the struct sizes in; that was quite an evil thing to do.
		// take 1 MB for headers/padding/etc (we don't need the full 1 MB but it's easier to just guess...)
		bspsize = 1024 * 1024;

		bspsize += numplanes * sizeof (dplane_t);
		bspsize += numleafs * sizeof (dleaf_t);
		bspsize += numvertexes * sizeof (dvertex_t);
		bspsize += numnodes * sizeof (dnode_t);
		bspsize += numtexinfo * sizeof (texinfo_t);
		bspsize += numfaces * sizeof (dface_t);
		bspsize += numclipnodes * sizeof (dclipnode_t);
		bspsize += nummarksurfaces * sizeof (int);
		bspsize += numsurfedges * sizeof (int);
		bspsize += numedges * sizeof (dedge_t);
		bspsize += nummodels * sizeof (dmodel_t);
		bspsize += lightdatasize;
		bspsize += visdatasize;
		bspsize += entdatasize;
		bspsize += texdatasize;

		SB_Alloc (&sb, bspsize);
		printf ("Allocated %f MB (%d bytes) for file buffer\n", bspsize* (1.0f / (1024.0f * 1024.0f)), bspsize);

		// write header
		SB_WriteInt (&sb, BSPVERSION);

		SB_ZeroFill (&sb, sizeof (lumps));	// filled in later

		SB_WriteLump (&sb, &lumps[LUMP_PLANES], dplanes, numplanes * sizeof (dplane_t));
		SB_WriteLump (&sb, &lumps[LUMP_LEAFS], dleafs, numleafs * sizeof (dleaf_t));
		SB_WriteLump (&sb, &lumps[LUMP_VERTEXES], dvertexes, numvertexes * sizeof (dvertex_t));
		SB_WriteLump (&sb, &lumps[LUMP_NODES], dnodes, numnodes * sizeof (dnode_t));
		SB_WriteLump (&sb, &lumps[LUMP_TEXINFO], texinfo, numtexinfo * sizeof (texinfo_t));
		SB_WriteLump (&sb, &lumps[LUMP_FACES], dfaces, numfaces * sizeof (dface_t));
		SB_WriteLump (&sb, &lumps[LUMP_CLIPNODES], dclipnodes, numclipnodes * sizeof (dclipnode_t));
		SB_WriteLump (&sb, &lumps[LUMP_MARKSURFACES], dmarksurfaces, nummarksurfaces * sizeof (dmarksurfaces[0]));
		SB_WriteLump (&sb, &lumps[LUMP_SURFEDGES], dsurfedges, numsurfedges * sizeof (dsurfedges[0]));
		SB_WriteLump (&sb, &lumps[LUMP_EDGES], dedges, numedges * sizeof (dedge_t));
		SB_WriteLump (&sb, &lumps[LUMP_MODELS], dmodels, nummodels * sizeof (dmodel_t));
		SB_WriteLump (&sb, &lumps[LUMP_LIGHTING], dlightdata, lightdatasize);
		SB_WriteLump (&sb, &lumps[LUMP_VISIBILITY], dvisdata, visdatasize);
		SB_WriteLump (&sb, &lumps[LUMP_ENTITIES], dentdata, entdatasize);
		SB_WriteLump (&sb, &lumps[LUMP_TEXTURES], dtexdata, texdatasize);

		// go back and update the header
		index = SB_Tell (&sb);
		SB_SeekAbsolute (&sb, 0);

		SB_WriteInt (&sb, BSPVERSION);

		for (i = 0; i < HEADER_LUMPS; i++)
		{
			SB_WriteInt (&sb, lumps[i].fileofs);
			SB_WriteInt (&sb, lumps[i].filelen);
		}

		SB_SeekAbsolute (&sb, index);

		// open a file and dump the buffer into it
		f = SafeOpenWrite (filename);
		SafeWrite (f, sb.start, (sb.index - sb.start));
		fclose (f);

		// finish up
		SB_Free (&sb);
	}

	if (rgblightdatasize)
	{
		FILE *litfile;
		litfile = SafeOpenWrite (filename_lit);

		if (litfile)
		{
			fputc ('Q', litfile);
			fputc ('L', litfile);
			fputc ('I', litfile);
			fputc ('T', litfile);
			fputc (1, litfile);
			fputc (0, litfile);
			fputc (0, litfile);
			fputc (0, litfile);
			SafeWrite (litfile, drgblightdata, rgblightdatasize);
			fclose (litfile);
		}
		else
			printf ("unable to write \"%s\"\n", filename_lit);
	}

	if (nmaplightdatasize)
	{
		FILE *litfile;
		litfile = SafeOpenWrite (filename_dlit);

		if (litfile)
		{
			fputc ('Q', litfile);
			fputc ('L', litfile);
			fputc ('I', litfile);
			fputc ('T', litfile);
			fputc (1, litfile);
			fputc (0, litfile);
			fputc (0, litfile);
			fputc (0, litfile);
			SafeWrite (litfile, dnmaplightdata, nmaplightdatasize);
			fclose (litfile);
		}
		else
			printf ("unable to write \"%s\"\n", filename_dlit);
	}
}

//============================================================================

/*
=============
PrintBSPFileSizes

Dumps info about current file
=============
*/
void PrintBSPFileSizes (void)
{
	printf ("%5i models       %6i\n", nummodels, (int) (nummodels * sizeof (dmodel_t)));
	printf ("%5i planes       %6i\n", numplanes, (int) (numplanes * sizeof (dplane_t)));
	printf ("%5i vertexes     %6i\n", numvertexes, (int) (numvertexes * sizeof (dvertex_t)));
	printf ("%5i nodes        %6i\n", numnodes, (int) (numnodes * sizeof (dnode_t)));
	printf ("%5i texinfo      %6i\n", numtexinfo, (int) (numtexinfo * sizeof (texinfo_t)));
	printf ("%5i faces        %6i\n", numfaces, (int) (numfaces * sizeof (dface_t)));
	printf ("%5i clipnodes    %6i\n", numclipnodes, (int) (numclipnodes * sizeof (dclipnode_t)));
	printf ("%5i leafs        %6i\n", numleafs, (int) (numleafs * sizeof (dleaf_t)));
	printf ("%5i marksurfaces %6i\n", nummarksurfaces, (int) (nummarksurfaces * sizeof (dmarksurfaces[0])));
	printf ("%5i surfedges    %6i\n", numsurfedges, (int) (numsurfedges * sizeof (dmarksurfaces[0])));
	printf ("%5i edges        %6i\n", numedges, (int) (numedges * sizeof (dedge_t)));

	if (!texdatasize)
		printf ("    0 textures          0\n");
	else
		printf ("%5i textures     %6i\n", ((dmiptexlump_t *) dtexdata)->nummiptex, texdatasize);

	printf ("      lightdata    %6i\n", lightdatasize);
	printf ("      visdata      %6i\n", visdatasize);
	printf ("      entdata      %6i\n", entdatasize);
}

//============================================================================

/*
===============
CompressVis
===============
*/
int CompressVis (byte *vis, byte *dest, int visrow)
{
	int		j;
	int		rep;
	byte	*dest_p;

	dest_p = dest;

	for (j = 0; j < visrow; j++)
	{
		*dest_p++ = vis[j];

		if (vis[j])
			continue;

		rep = 1;

		for (j++; j < visrow; j++)
			if (vis[j] || rep == 255)
				break;
			else
				rep++;

		*dest_p++ = rep;
		j--;
	}

	return dest_p - dest;
}

/*
===============
DecompressVis
===============
*/
void DecompressVis (byte *in, byte *out, int size)
{
	byte *end = out + size;
	int n;

	while (out < end)
	{
		n = *in++;

		if (n)
			*out++ = n;
		else
		{
			n = *in++;

			while (n--)
				*out++ = 0;
		}
	}
}

//============================================================================

int			num_entities = 0;
entity_t	entities[MAX_MAP_ENTITIES];

void PrintEntity (entity_t *ent)
{
	epair_t	*ep;

	for (ep = ent->epairs; ep; ep = ep->next)
		printf ("%20s : %s\n", ep->key, ep->value);
}


char *ValueForKey (entity_t *ent, char *key)
{
	epair_t	*ep;

	for (ep = ent->epairs; ep; ep = ep->next)
		if (!strcmp (ep->key, key))
			return ep->value;

	return "";
}

vec_t FloatForKey (entity_t *ent, char *key)
{
	epair_t	*ep;

	for (ep = ent->epairs; ep; ep = ep->next)
		if (!strcmp (ep->key, key))
			return atof (ep->value);

	return 0;
}

void SetKeyValue (entity_t *ent, char *key, char *value)
{
	epair_t	*ep;

	for (ep = ent->epairs; ep; ep = ep->next)
		if (!strcmp (ep->key, key))
		{
			qfree (ep->value);
			ep->value = copystring (value);
			return;
		}

	ep = qmalloc (sizeof (*ep));
	ep->next = ent->epairs;
	ent->epairs = ep;
	ep->key = copystring (key);
	ep->value = copystring (value);
}

/*
=================
ParseEpair
=================
*/
epair_t *ParseEpair (void)
{
	epair_t	*e;

	e = qmalloc (sizeof (epair_t));
	memset (e, 0, sizeof (epair_t));

	if (strlen (token) >= MAX_KEY - 1)
		Error ("ParseEpair: token too long");

	e->key = copystring (token);
	GetToken (false);

	if (strlen (token) >= MAX_VALUE - 1)
		Error ("ParseEpair: token too long");

	e->value = copystring (token);

	return e;
}

entity_t *FindEntityWithKeyPair (char *key, char *value)
{
	entity_t *ent;
	epair_t	*ep;
	int i;

	for (i = 0; i < num_entities; i++)
	{
		ent = &entities[ i ];

		for (ep = ent->epairs; ep; ep = ep->next)
		{
			if (!strcmp (ep->key, key))
			{
				if (!strcmp (ep->value, value))
					return ent;

				break;
			}
		}
	}

	return NULL;
}

void GetVectorForKey (entity_t *ent, char *key, vec3_t vec)
{
	char	*k;
	double	v1, v2, v3;

	k = ValueForKey (ent, key);
	v1 = v2 = v3 = 0;
	// scanf into doubles, then assign, so it is vec_t size independent
	sscanf (k, "%lf %lf %lf", &v1, &v2, &v3);
	vec[0] = v1;
	vec[1] = v2;
	vec[2] = v3;
}

/*
==================
ParseEntities
==================
*/
void ParseEntities (void)
{
	char 		*data;
	entity_t	*entity;
	char		key[MAX_KEY];
	epair_t		*epair;

	data = dentdata;

	//
	// start parsing
	//
	num_entities = 0;

	// go through all the entities
	while (1)
	{
		// parse the opening brace
		data = COM_Parse (data);

		if (!data)
			break;

		if (com_token[0] != '{')
			Error ("LoadEntities: found %s when expecting {", com_token);

		if (num_entities == MAX_MAP_ENTITIES)
			Error ("LoadEntities: MAX_MAP_ENTITIES");

		entity = &entities[num_entities++];

		// go through all the keys in this entity
		while (1)
		{
			int		c;

			// parse key
			data = COM_Parse (data);

			if (!data)
				Error ("LoadEntities: EOF without closing brace");

			if (!strcmp (com_token, "}"))
				break;

			strcpy (key, com_token);

			// parse value
			data = COM_Parse (data);

			if (!data)
				Error ("LoadEntities: EOF without closing brace");

			c = com_token[0];

			if (c == '}')
				Error ("LoadEntities: closing brace without data");

			epair = malloc (sizeof (epair_t));
			epair->key = copystring (key);
			epair->value = copystring (com_token);
			epair->next = entity->epairs;
			entity->epairs = epair;
		}
	}
}

/*
==================
UnparseEntity
==================
*/
static void UnparseEntity (entity_t *ent, char *buf, char **end, qboolean ws)
{
	epair_t		*ep;
	char		line[16384];

	ep = ent->epairs;

	if (!ep)
		return;

	strcat (*end, "{\n");
	*end += 2;

	for (; ep; ep = ep->next)
	{
		if (ws && (!strcmp (ep->key, "angle") || !strcmp (ep->key, "angles")))
			continue;

		sprintf (line, "\"%s\" \"%s\"\n", ep->key, ep->value);
		strcat (*end, line);
		*end += strlen (line);
	}

	strcat (*end, "}\n");
	*end += 2;

	if (*end > buf + MAX_MAP_ENTSTRING)
		Error ("Entity text too long");
}

/*
==================
UnparseEntities
==================
*/
void UnparseEntities (void)
{
	int			i;
	entity_t	*ent;
	char		*buf, *end;

	buf = dentdata;
	end = buf;
	*end = 0;

	// Vic: write "worldspawn" as the very first entity (engines might depend on it)
	ent = FindEntityWithKeyPair ("classname", "worldspawn");

	if (ent)
		UnparseEntity (ent, buf, &end, true);

	for (i = 0; i < num_entities; i++)
	{
		if (&entities[i] != ent)
			UnparseEntity (&entities[i], buf, &end, false);
	}

	entdatasize = end - buf + 1;
}
