/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// r_alias.c -- alias model rendering
// this file mostly just handles IQM now

#include "quakedef.h"

#define BLENDIQMPOSE \
	"TEMP transvert;\n" \
	"DPH transvert.x, vertex.attrib[0], vertex.attrib[3];\n" \
	"DPH transvert.y, vertex.attrib[0], vertex.attrib[4];\n" \
	"DPH transvert.z, vertex.attrib[0], vertex.attrib[5];\n" \
	"MOV transvert.w, vertex.attrib[0].w;\n" \
	TRANSFORMVERTEX ("result.position", "transvert")

// this lighting is not correct per OpenGL's lighting model but it is consistent with MDLs
#define iqm_vp \
	"!!ARBvp1.0\n" \
	"TEMP transnorm, dot, dotlow, dothigh;\n" \
	BLENDIQMPOSE \
	"DP3 transnorm.x, vertex.attrib[1], vertex.attrib[3];\n" \
	"DP3 transnorm.y, vertex.attrib[1], vertex.attrib[4];\n" \
	"DP3 transnorm.z, vertex.attrib[1], vertex.attrib[5];\n" \
	"MUL result.texcoord[0], vertex.attrib[2], program.local[0];\n" \
	"MUL result.texcoord[1], vertex.attrib[2], program.local[1];\n" \
	"DP3 dot, transnorm, program.env[0];\n" \
	"ADD dothigh, dot, 1.0;\n" \
	"MAD dotlow, dot, 0.2954545, 1.0;\n" \
	"MAX result.texcoord[2], dotlow, dothigh;\n" \
	"END\n"

#define iqm_shadow_vp \
	"!!ARBvp1.0\n" \
	BLENDIQMPOSE \
	"END\n"

#define iqm_fp \
	"!!ARBfp1.0\n" \
	"TEMP light, tex0;\n" \
	"MUL light, fragment.texcoord[2], program.env[1]; # shadelight\n" \
	"TEX tex0, fragment.texcoord[0], texture[0], 2D;\n" \
	"MUL result.color, tex0, light;\n" \
	"MOV result.color.w, program.env[1].w; # alpha\n" \
	"END\n"

#define iqm_fp_fb \
	"!!ARBfp1.0\n" \
	"TEMP light, tex0, tex1;\n" \
	"MUL light, fragment.texcoord[2], program.env[1]; # shadelight\n" \
	"TEX tex0, fragment.texcoord[0], texture[0], 2D;\n" \
	"TEX tex1, fragment.texcoord[1], texture[1], 2D;\n" \
	FULLBRIGHTBLEND ("result.color", "tex0", "light", "tex1") \
	"MOV result.color.w, program.env[1].w; # alpha\n" \
	"END\n"

#define iqm_shadow_fp \
	"!!ARBfp1.0\n" \
	"MOV result.color, program.local[0];\n" \
	"END\n"

GLuint arb_iqm_vp = 0;
GLuint arb_iqm_fp = 0;
GLuint arb_iqm_fp_fog = 0;
GLuint arb_iqm_fp_fb = 0;
GLuint arb_iqm_fp_fb_fog = 0;
GLuint arb_iqm_shadow_vp = 0;
GLuint arb_iqm_shadow_fp = 0;
GLuint arb_iqm_shadow_fp_fog = 0;

void IQM_CreateShaders (void)
{
	arb_iqm_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, iqm_vp, false);
	arb_iqm_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, iqm_fp, false);
	arb_iqm_fp_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, iqm_fp, true);
	arb_iqm_fp_fb = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, iqm_fp_fb, false);
	arb_iqm_fp_fb_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, iqm_fp_fb, true);
	arb_iqm_shadow_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, iqm_shadow_vp, false);
	arb_iqm_shadow_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, iqm_shadow_fp, false);
	arb_iqm_shadow_fp_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, iqm_shadow_fp, true);
}

void GLARB_PadAliasTMU (gltexture_t *tx, int local);
void GLARB_SetupAliasLighting (entity_t *e, aliasstate_t *state);

GLuint r_iqmvbo = 0;
GLuint r_iqmindexes = 0;

extern cvar_t r_drawflat, gl_fullbrights, r_lerpmodels, r_lerpmove; // johnfitz

// up to 16 color translated skins
gltexture_t *playertextures[MAX_SCOREBOARD]; // johnfitz -- changed to an array of pointers
extern unsigned int r_meshindexbuffer;
extern unsigned int r_meshvertexbuffer;

extern	vec3_t lightspot;


void GLARB_SetupAliasModel (entity_t *e);
void GLARB_DrawAliasBatches (entity_t **ents, int numents);
void GLARB_DrawShadowBatches (entity_t **ents, int numents);


/*
=================
R_SetupAliasFrame -- johnfitz -- rewritten to support lerping
=================
*/
void R_SetupAliasFrame (entity_t *e, aliashdr_t *paliashdr, int frame, aliasstate_t *state)
{
}

void R_MinimumLighting (float *lightcolor, float minimum)
{
	float add = minimum - (lightcolor[0] + lightcolor[1] + lightcolor[2]);

	if (add > 0.0f)
	{
		lightcolor[0] += add / 3.0f;
		lightcolor[1] += add / 3.0f;
		lightcolor[2] += add / 3.0f;
	}
}


void R_DrawAliasShadow (entity_t *ent, aliashdr_t *hdr, aliasstate_t *state, float *dstmesh, float *shadecolor)
{
}


void R_DrawAliasShadows (entity_t **ents, int numents, void *meshbuffer)
{
}


void R_SetupAliasModel (entity_t *e)
{
}


void R_DrawAliasModel (entity_t *ent, aliashdr_t *hdr, aliasstate_t *state, qboolean showtris)
{
}


void R_DrawAliasBatchPass (entity_t **ents, int numents, qboolean showtris)
{
}


void R_DrawAliasBatches (entity_t **ents, int numents, void *meshbuffer)
{
}


int R_DrawAliasModels_Sort (entity_t **e1, entity_t **e2)
{
	// sort by texture (the same model can have different textures)
	if (e1[0]->aliasstate.tx == e2[0]->aliasstate.tx)
		return e1[0]->encodedpose - e2[0]->encodedpose;
	else return (int) (e1[0]->aliasstate.tx - e2[0]->aliasstate.tx);
}


void R_AddEntityToAlphaList (entity_t *ent);

void R_DrawAliasModels (void)
{
	int i;
	int numalias;

	// temp buffer for storing ents for sorting
	entity_t **ents = (entity_t **) scratchbuf;

	if (!r_drawentities.value) return;

	for (i = 0, numalias = 0; i < cl_numvisedicts; i++)
	{
		entity_t *ent = cl_visedicts[i];

		if (!ent->model) continue;
		if (ent->model->type != mod_alias) continue;

		if (ent->alpha < 1)
			ent->alpha = 255;
		else if (ent->alpha < 255)
		{
			R_AddEntityToAlphaList (ent);
			continue;
		}

		GLARB_SetupAliasModel (ent);

		// if it was culled or otherwise removed from the renderer we don't bother with it
		// (unless it's got shadows enabled in which case we have to :( )
		if (ent->visframe != r_framecount && r_shadows.value < 0.01f) continue;

		// store this model
		ents[numalias++] = ent;
	}

	if (!numalias) return;

	// sort the models by texture so that we can batch the bastards
	qsort (ents, numalias, sizeof (entity_t *), (sortfunc_t) R_DrawAliasModels_Sort);

	GLARB_DrawAliasBatches (ents, numalias);
	GLARB_DrawShadowBatches (ents, numalias);
}


// valiant attempt at clawing back some CPU from IQM animation #1
void Matrix3x4_ScaleAdd (matrix3x4_t *out, matrix3x4_t *base, float scale, matrix3x4_t *add)
{
	out->a[0] = base->a[0] * scale + add->a[0];
	out->a[1] = base->a[1] * scale + add->a[1];
	out->a[2] = base->a[2] * scale + add->a[2];
	out->a[3] = base->a[3] * scale + add->a[3];

	out->b[0] = base->b[0] * scale + add->b[0];
	out->b[1] = base->b[1] * scale + add->b[1];
	out->b[2] = base->b[2] * scale + add->b[2];
	out->b[3] = base->b[3] * scale + add->b[3];

	out->c[0] = base->c[0] * scale + add->c[0];
	out->c[1] = base->c[1] * scale + add->c[1];
	out->c[2] = base->c[2] * scale + add->c[2];
	out->c[3] = base->c[3] * scale + add->c[3];
}


// valiant attempt at clawing back some CPU from IQM animation #2
void Matrix3x4_ScaleScaleAdd (matrix3x4_t *out, matrix3x4_t *base1, float scale1, matrix3x4_t *base2, float scale2)
{
	out->a[0] = ((base2->a[0] - base1->a[0]) * scale2) + base1->a[0];
	out->a[1] = ((base2->a[1] - base1->a[1]) * scale2) + base1->a[1];
	out->a[2] = ((base2->a[2] - base1->a[2]) * scale2) + base1->a[2];
	out->a[3] = ((base2->a[3] - base1->a[3]) * scale2) + base1->a[3];

	out->b[0] = ((base2->b[0] - base1->b[0]) * scale2) + base1->b[0];
	out->b[1] = ((base2->b[1] - base1->b[1]) * scale2) + base1->b[1];
	out->b[2] = ((base2->b[2] - base1->b[2]) * scale2) + base1->b[2];
	out->b[3] = ((base2->b[3] - base1->b[3]) * scale2) + base1->b[3];

	out->c[0] = ((base2->c[0] - base1->c[0]) * scale2) + base1->c[0];
	out->c[1] = ((base2->c[1] - base1->c[1]) * scale2) + base1->c[1];
	out->c[2] = ((base2->c[2] - base1->c[2]) * scale2) + base1->c[2];
	out->c[3] = ((base2->c[3] - base1->c[3]) * scale2) + base1->c[3];
}


// valiant attempt at clawing back some CPU from IQM animation #3
void Matrix3x4_MultiplyByRef (matrix3x4_t *out, matrix3x4_t *mat1, matrix3x4_t *mat2)
{
	// this was massively cutesy but slower than it should be; it's already slow, let's not make it worse
	out->a[0] = (mat2->a[0] * mat1->a[0]) + (mat2->b[0] * mat1->a[1]) + (mat2->c[0] * mat1->a[2]);
	out->a[1] = (mat2->a[1] * mat1->a[0]) + (mat2->b[1] * mat1->a[1]) + (mat2->c[1] * mat1->a[2]);
	out->a[2] = (mat2->a[2] * mat1->a[0]) + (mat2->b[2] * mat1->a[1]) + (mat2->c[2] * mat1->a[2]);
	out->a[3] = (mat2->a[3] * mat1->a[0]) + (mat2->b[3] * mat1->a[1]) + (mat2->c[3] * mat1->a[2]) + mat1->a[3];

	out->b[0] = (mat2->a[0] * mat1->b[0]) + (mat2->b[0] * mat1->b[1]) + (mat2->c[0] * mat1->b[2]);
	out->b[1] = (mat2->a[1] * mat1->b[0]) + (mat2->b[1] * mat1->b[1]) + (mat2->c[1] * mat1->b[2]);
	out->b[2] = (mat2->a[2] * mat1->b[0]) + (mat2->b[2] * mat1->b[1]) + (mat2->c[2] * mat1->b[2]);
	out->b[3] = (mat2->a[3] * mat1->b[0]) + (mat2->b[3] * mat1->b[1]) + (mat2->c[3] * mat1->b[2]) + mat1->b[3];

	out->c[0] = (mat2->a[0] * mat1->c[0]) + (mat2->b[0] * mat1->c[1]) + (mat2->c[0] * mat1->c[2]);
	out->c[1] = (mat2->a[1] * mat1->c[0]) + (mat2->b[1] * mat1->c[1]) + (mat2->c[1] * mat1->c[2]);
	out->c[2] = (mat2->a[2] * mat1->c[0]) + (mat2->b[2] * mat1->c[1]) + (mat2->c[2] * mat1->c[2]);
	out->c[3] = (mat2->a[3] * mat1->c[0]) + (mat2->b[3] * mat1->c[1]) + (mat2->c[3] * mat1->c[2]) + mat1->c[3];
}


void GL_AnimateIQMFrame (iqmdata_t *iqm, int currframe, int lastframe, float lerp)
{
	int i;

    int frame1 = lastframe;
    int frame2 = currframe;

	matrix3x4_t *destmatrix = (matrix3x4_t *) scratchbuf;

	// frame sanity - prevent from going out of bounds
	frame1 %= iqm->num_poses;
	frame2 %= iqm->num_poses;

	// test for flipping the frames
	if (frame1 == iqm->cachedcurrframe && frame2 == iqm->cachedlastframe)
	{
		int temp = frame1;

		frame1 = frame2;
		frame2 = temp;
		lerp = 1.0f - lerp;
	}

	// coarsen the lerp so that we don't need to recache too often
	// this animates at 50 FPS which should be enough for anyone
	lerp = (float) ((int) (lerp * 5)) / 5;

	// if the cached frames and lerp don't change there is no need to re-animate
	// these are all that it's safe to cache as the transformed bones are stored in the scratchbuf
	// we can only cache per model too as the data set is just too big to cache per entity
	if (frame1 != iqm->cachedlastframe || frame2 != iqm->cachedcurrframe || lerp != iqm->cachedlerp)
	{
		matrix3x4_t *mat1 = &iqm->frames[frame1 * iqm->num_joints];
		matrix3x4_t *mat2 = &iqm->frames[frame2 * iqm->num_joints];
		matrix3x4_t mat;

		if (iqm->version == IQM_VERSION1)
		{
			for (i = 0; i < iqm->num_joints; i++)
			{
				Matrix3x4_ScaleScaleAdd (&mat, &mat1[i], (1 - lerp), &mat2[i], lerp);

				if (iqm->jointsv1[i].parent >= 0)
					Matrix3x4_MultiplyByRef (&iqm->outframe[i], &iqm->outframe[iqm->jointsv1[i].parent], &mat);
				else Matrix3x4_Copy (&iqm->outframe[i], &mat);
			}
		}
		else
		{
			for (i = 0; i < iqm->num_joints; i++)
			{
				Matrix3x4_ScaleScaleAdd (&mat, &mat1[i], (1 - lerp), &mat2[i], lerp);

				if (iqm->jointsv2[i].parent >= 0)
					Matrix3x4_MultiplyByRef (&iqm->outframe[i], &iqm->outframe[iqm->jointsv2[i].parent], &mat);
				else Matrix3x4_Copy (&iqm->outframe[i], &mat);
			}
		}
	}

	// cache back the frames
	iqm->cachedlastframe = frame1;
	iqm->cachedcurrframe = frame2;
	iqm->cachedlerp = lerp;

	// The actual vertex generation based on the matrixes follows...
	{
		// blendweights were converted to float on load; this consumes maybe an extra 20k memory
		// for a big model but runs a coupla percent faster - that's the tradeoff
		const unsigned char *index = iqm->blendindexes;
		const float *weight = iqm->blendweights;

		for (i = 0; i < iqm->numvertexes; i++, index += 4, weight += 4)
		{
			Matrix3x4_Scale (&destmatrix[i], &iqm->outframe[index[0]], weight[0]);

			// yet another valiant attempt at clawing back some CPU from IQM animation
			if (weight[1])
			{
				Matrix3x4_ScaleAdd (&destmatrix[i], &iqm->outframe[index[1]], weight[1], &destmatrix[i]);

				if (weight[2])
				{
					Matrix3x4_ScaleAdd (&destmatrix[i], &iqm->outframe[index[2]], weight[2], &destmatrix[i]);

					if (weight[3])
					{
						Matrix3x4_ScaleAdd (&destmatrix[i], &iqm->outframe[index[3]], weight[3], &destmatrix[i]);
					}
				}
			}
		}
	}
}


void R_SetupIQMFrame (entity_t *e, iqmdata_t *iqm, int frame, aliasstate_t *state)
{
	int posenum;

	if ((frame >= iqm->num_poses) || (frame < 0))
	{
//		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	posenum = frame;
	e->lerptime = 0.1;

	if (e->lerpflags & LERP_RESETANIM) // kill any lerp in progress
	{
		e->lerpstart = 0;
		e->previouspose = posenum;
		e->currentpose = posenum;
		e->lerpflags -= LERP_RESETANIM;
	}
	else if (e->currentpose != posenum) // pose changed, start new lerp
	{
		if (e->lerpflags & LERP_RESETANIM2) // defer lerping one more time
		{
			e->lerpstart = 0;
			e->previouspose = posenum;
			e->currentpose = posenum;
			e->lerpflags -= LERP_RESETANIM2;
		}
		else
		{
			e->lerpstart = cl.time;
			e->previouspose = e->currentpose;
			e->currentpose = posenum;
		}
	}

	// set up values
	if (r_lerpmodels.value && !((e->model->flags & MOD_NOLERP) && r_lerpmodels.value != 2))
	{
		if (e->lerpflags & LERP_FINISH)
			state->blend = CLAMP (0, (cl.time - e->lerpstart) / (e->lerpfinish - e->lerpstart), 1);
		else state->blend = CLAMP (0, (cl.time - e->lerpstart) / e->lerptime, 1);

		state->prevpose = e->previouspose;
		state->currpose = e->currentpose;
	}
	else // don't lerp
	{
		state->blend = 1;
		state->prevpose = posenum;
		state->currpose = posenum;
	}
}


void GLIQM_Transform (entity_t *ent, iqmdata_t *iqm)
{
	GL_TranslateMatrix (&ent->matrix, ent->origin[0], ent->origin[1], ent->origin[2]);

	if (ent->angles[1]) GL_RotateMatrix (&ent->matrix, ent->angles[1], 0, 0, 1);
	if (ent->angles[0]) GL_RotateMatrix (&ent->matrix, -ent->angles[0], 0, 1, 0);
	if (ent->angles[2]) GL_RotateMatrix (&ent->matrix, ent->angles[2], 1, 0, 0);

	// DP_ENT_SCALE
	if (ent->scale != 16)
		GL_ScaleMatrix (&ent->matrix, (float) ent->scale / 16.0f, (float) ent->scale / 16.0f, (float) ent->scale / 16.0f);

#ifdef RMQMSVC
	// crap heuristic to detect mrfixit for testing so that we can rescale it and have it appear normally in-game
	// remove this for production!!!!!
	if (iqm->num_joints == 75 && iqm->radius > 8.1 && iqm->radius < 8.2 && iqm->num_poses == 101 && iqm->num_triangles == 2988 && iqm->numvertexes == 1861)
	{
		GL_ScaleMatrix (&ent->matrix, 8, 8, 8);
		GL_TranslateMatrix (&ent->matrix, 0, 0, -3);
	}
#endif
}


void R_DrawIQMModel (entity_t *ent, model_t *mod, iqmdata_t *iqm)
{
	int i;
	aliasstate_t *state = &ent->aliasstate;
	extern int gl_stencilbits;
	extern glmatrix *r_shadowmatrix;
	float an, lightvec[4];

	// we could do this in our vertex shader but it would require passing the normal too, as well as
	// some branching, so for the target hardware this is quite sufficient.
	an = (state->angles[1] + state->angles[0]) / 180 * M_PI;
	lightvec[0] = cos (-an);
	lightvec[1] = sin (-an);
	lightvec[2] = 1;
	VectorNormalize (lightvec);

#ifdef RMQMSVC
	// crap heuristic to detect mrfixit for testing so that we can animate it
	// remove this for production!!!!!
	if (iqm->num_joints == 75 && iqm->radius > 8.1 && iqm->radius < 8.2 && iqm->num_poses == 101 && iqm->num_triangles == 2988 && iqm->numvertexes == 1861)
	{
		state->currpose = ceil (cl.time * 10);
		state->prevpose = floor (cl.time * 10);
		state->blend = (cl.time * 10) - state->prevpose;
	}
#endif

	// animate the model (this may be skipped if the frames haven't changed since last time this model was animated)
	GL_AnimateIQMFrame (iqm, state->currpose, state->prevpose, state->blend);

	// set up lighting
	GLARB_SetupAliasLighting (ent, &ent->aliasstate);

	// these need to become env params because the shaders have not been set yet
	qglProgramEnvParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, lightvec);
	qglProgramEnvParameter4fvARB (GL_FRAGMENT_PROGRAM_ARB, 1, ent->aliasstate.shadelight);

	if (ent->efrag && !(ent->model->flags & EF_ROTATE))
	{
		glLoadMatrixf (r_world_matrix.m16);
		glMultMatrixf (ent->matrix.m16);
	}
	else
	{
		// go back to the world matrix
		GL_LoadMatrix (&ent->matrix, &r_world_matrix);
		GLIQM_Transform (ent, iqm);
		glLoadMatrixf (ent->matrix.m16);
	}

	// note - the iqm can go in a static vertex buffer now... ;)
	GL_VertexArrayPointer (r_iqmvbo, 0, 3, GL_FLOAT, GL_FALSE, sizeof (iqmvertex_t), (void *) iqm->vertexoffset);
	GL_VertexArrayPointer (r_iqmvbo, 1, 3, GL_FLOAT, GL_FALSE, sizeof (iqmvertex_t), (void *) (iqm->vertexoffset + 12));
	GL_VertexArrayPointer (r_iqmvbo, 2, 2, GL_FLOAT, GL_FALSE, sizeof (iqmvertex_t), (void *) (iqm->vertexoffset + 24));

	GL_SetVertexProgram (arb_iqm_vp);

	for (i = 0; i < iqm->num_mesh; i++)
	{
		iqmmesh_t *mesh = &iqm->mesh[i];

		// don't bother drawing it if a skin wasn't loaded
		if (iqm->skins[i])
		{
			GLARB_PadAliasTMU (iqm->skins[i], 0);
			GL_BindTexture (GL_TEXTURE0, iqm->skins[i]);

			if (iqm->fullbrights[i])
			{
				GLARB_PadAliasTMU (iqm->fullbrights[i], 1);
				GL_BindTexture (GL_TEXTURE1, iqm->fullbrights[i]);

				if (Fog_GetDensity () > 0)
					GL_SetFragmentProgram (arb_iqm_fp_fb_fog);
				else GL_SetFragmentProgram (arb_iqm_fp_fb);
			}
			else
			{
				if (Fog_GetDensity () > 0)
					GL_SetFragmentProgram (arb_iqm_fp_fog);
				else GL_SetFragmentProgram (arb_iqm_fp);
			}

			GL_DrawElements (r_iqmindexes,
				GL_TRIANGLES,
				mesh->first_vertex,
				mesh->first_vertex + mesh->num_vertexes - 1,
				3 * mesh->num_triangles,
				GL_UNSIGNED_SHORT,
				(void *) iqm->indexoffsets[i]);

			rs_aliaspolys += mesh->num_triangles;	// consistent with ID Quake
		}
	}

	// no shadow
	if (r_shadows.value <= 0) return;
	if (ent == &cl.viewent) return;
	if (ent->model->flags & MOD_NOSHADOW) return;
	if (ent->alpha < 1) ent->alpha = 255;
	if (ent->alpha < 255) return;

	// eeeewwww.  shadows begin here.  running the CPU animation a second time is a major speed drain so we
	// just suck up the state changes instead.  only alternative is to cache the animation in memory which
	// would be prohibitively expensive (if you're thinking "this sucks" you're probably right).
	if (gl_stencilbits)
	{
		glEnable (GL_STENCIL_TEST);
		glStencilFunc (GL_EQUAL, 1, 2);
		glStencilOp (GL_KEEP, GL_KEEP, GL_INCR);
	}

	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);

	GL_SetVertexProgram (arb_iqm_shadow_vp);

	if (Fog_GetDensity () > 0)
		GL_SetFragmentProgram (arb_iqm_shadow_fp_fog);
	else GL_SetFragmentProgram (arb_iqm_shadow_fp);

	qglProgramLocalParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 0, 0, 0, 0, (r_shadows.value < 1 ? r_shadows.value : 1));

	// go back to the world matrix
	GL_LoadMatrix (&ent->matrix, &r_world_matrix);

	GL_TranslateMatrix (&ent->matrix, ent->origin[0], ent->origin[1], ent->origin[2]);

	GL_TranslateMatrix (&ent->matrix, 0, 0, -(state->origin[2] - state->lightspot[2]));
	GL_MultiplyMatrix (&ent->matrix, r_shadowmatrix, &ent->matrix);
	GL_TranslateMatrix (&ent->matrix, 0, 0, (state->origin[2] - state->lightspot[2]));

	GL_RotateMatrix (&ent->matrix, ent->angles[1], 0, 0, 1);

	// DP_ENT_SCALE
	if (ent->scale != 16)
		GL_ScaleMatrix (&ent->matrix, (float) ent->scale / 16.0f, (float) ent->scale / 16.0f, (float) ent->scale / 16.0f);

#ifdef RMQMSVC
	// crap heuristic to detect mrfixit for testing so that we can rescale it and have it appear normally in-game
	// remove this for production!!!!!
	if (iqm->num_joints == 75 && iqm->radius > 8.1 && iqm->radius < 8.2 && iqm->num_poses == 101 && iqm->num_triangles == 2988 && iqm->numvertexes == 1861)
	{
		GL_ScaleMatrix (&ent->matrix, 8, 8, 8);
		GL_TranslateMatrix (&ent->matrix, 0, 0, -3);
	}
#endif

	glLoadMatrixf (ent->matrix.m16);

	for (i = 0; i < iqm->num_mesh; i++)
	{
		iqmmesh_t *mesh = &iqm->mesh[i];

		// don't bother drawing it if a skin wasn't loaded
		if (iqm->skins[i])
		{
			GL_DrawElements (r_iqmindexes,
				GL_TRIANGLES,
				mesh->first_vertex,
				mesh->first_vertex + mesh->num_vertexes - 1,
				3 * mesh->num_triangles,
				GL_UNSIGNED_SHORT,
				(void *) iqm->indexoffsets[i]);

			rs_aliaspolys += mesh->num_triangles;	// consistent with ID Quake
		}
	}

	if (gl_stencilbits)
		glDisable (GL_STENCIL_TEST);

	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
}


void R_SetupIQMState (void)
{
	if (gl_smoothmodels.value) glShadeModel (GL_SMOOTH);
	if (gl_affinemodels.value) glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2 | VAA_3 | VAA_4 | VAA_5);

	// yuck
	GL_VertexArrayPointer (0, 3, 4, GL_FLOAT, GL_FALSE, sizeof (matrix3x4_t), ((matrix3x4_t *) scratchbuf)->a);
	GL_VertexArrayPointer (0, 4, 4, GL_FLOAT, GL_FALSE, sizeof (matrix3x4_t), ((matrix3x4_t *) scratchbuf)->b);
	GL_VertexArrayPointer (0, 5, 4, GL_FLOAT, GL_FALSE, sizeof (matrix3x4_t), ((matrix3x4_t *) scratchbuf)->c);
}


void R_TakedownIQMState (void)
{
	glLoadMatrixf (r_world_matrix.m16);

	if (gl_affinemodels.value) glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	if (gl_smoothmodels.value) glShadeModel (GL_FLAT);
}


void R_DrawIQMModels (void)
{
	int i;
	int numalias;
	qboolean stateset = false;

	if (!r_drawentities.value) return;

	for (i = 0, numalias = 0; i < cl_numvisedicts; i++)
	{
		entity_t *ent = cl_visedicts[i];

		if (!ent->model) continue;
		if (ent->model->type != mod_iqm) continue;

		if (R_CullBox (ent->mins, ent->maxs)) continue;

		ent->visframe = r_framecount;

		if (ent->alpha < 1)
			ent->alpha = 255;
		else if (ent->alpha < 255)
		{
			R_AddEntityToAlphaList (ent);
			continue;
		}

		if (!stateset)
		{
			R_SetupIQMState ();
			stateset = true;
		}

		// draw the model
		R_DrawIQMModel (ent, ent->model, ent->model->iqmdata);
	}

	if (stateset) R_TakedownIQMState ();
}


/*
=============
R_DrawViewModel -- johnfitz -- gutted

mh - now goes through same draw routine as other mdls
=============
*/
void GLARB_SetupAliasLerp (entity_t *e, aliashdr_t *hdr);
void GLARB_LerpAliasMove (entity_t *e, aliasstate_t *state);

extern cvar_t scr_fov;
void GL_SetFrustum (glmatrix *matrix, float fovx, float fovy);
void SCR_CalcFOV (refdef_t *rd, float fov);

void R_DrawViewModel (void)
{
	entity_t *e = &cl.viewent;

	if (!r_drawviewmodel.value || !r_drawentities.value || chase_active.value || envmap) return;
	if (cl.stats[STAT_HEALTH] <= 0) return;

	if (!e->model) return;

	if (cl.items & IT_INVISIBILITY)
		e->alpha = 64;
	else e->alpha = 255;

	if (r_drawviewmodel.value < 1)
		e->alpha = (byte) (((float) e->alpha * r_drawviewmodel.value) + 0.5f);

	if (e->alpha < 1)
		return;
	else if (e->alpha < 255)
	{
		glEnable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	if (scr_fov.value > 90)
	{
		refdef_t vm_refdef;
		glmatrix newproj;

		glMatrixMode (GL_PROJECTION);
		glPushMatrix ();

		SCR_CalcFOV (&vm_refdef, 90);
		GL_SetFrustum (&newproj, vm_refdef.fov_x, vm_refdef.fov_y);

		glMatrixMode (GL_MODELVIEW);
	}

	if (e->model->type == mod_alias)
	{
		aliashdr_t *hdr = Mod_Extradata (e->model);

		e->visframe = r_framecount;
		GLARB_SetupAliasModel (e);
		GLARB_SetupAliasLerp (e, hdr);
		GLARB_LerpAliasMove (e, &e->aliasstate);
		GLARB_DrawAliasBatches (&e, 1);
	}
	else if (e->model->type == mod_iqm)
	{
		e->visframe = r_framecount;
		R_SetupIQMFrame (e, e->model->iqmdata, e->frame, &e->aliasstate);
		GLARB_LerpAliasMove (e, &e->aliasstate);
		R_SetupIQMState ();
		R_DrawIQMModel (e, e->model, e->model->iqmdata);
		R_TakedownIQMState ();
	}

	if (scr_fov.value > 90)
	{
		glMatrixMode (GL_PROJECTION);
		glPopMatrix ();
		glMatrixMode (GL_MODELVIEW);
	}

	if (e->alpha < 255)
	{
		glDisable (GL_BLEND);
	}
}


void GLIQM_LoadVertexBuffers (void)
{
	int i, j;
	model_t *m;
	int totalvbosize = 0;
	int totalindexes = 0;

	for (j = 1; j < MAX_MODELS; j++)
	{
		iqmdata_t *iqm;

		if (!(m = cl.model_precache[j])) break;
		if (m->type != mod_iqm) continue;

		iqm = m->iqmdata;

		iqm->vertexoffset = totalvbosize;
		totalvbosize += iqm->numvertexes * sizeof (iqmvertex_t);

		for (i = 0; i < iqm->num_mesh; i++)
		{
			iqmmesh_t *mesh = &iqm->mesh[i];

			iqm->indexoffsets[i] = totalindexes * sizeof (unsigned short);
			totalindexes += mesh->num_triangles * 3;
		}
	}

	if (!totalindexes) return;
	if (!totalvbosize) return;

	r_iqmindexes = GL_GetBuffer ();
	qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, r_iqmindexes);
	qglBufferData (GL_ELEMENT_ARRAY_BUFFER, totalindexes * sizeof (unsigned short), NULL, GL_STATIC_DRAW);

	r_iqmvbo = GL_GetBuffer ();
	qglBindBuffer (GL_ARRAY_BUFFER, r_iqmvbo);
	qglBufferData (GL_ARRAY_BUFFER, totalvbosize, NULL, GL_STATIC_DRAW);

	for (j = 1; j < MAX_MODELS; j++)
	{
		iqmdata_t *iqm;

		if (!(m = cl.model_precache[j])) break;
		if (m->type != mod_iqm) continue;

		iqm = m->iqmdata;

		qglBufferSubData (GL_ARRAY_BUFFER, iqm->vertexoffset, iqm->numvertexes * sizeof (iqmvertex_t), iqm->verts);

		for (i = 0; i < iqm->num_mesh; i++)
		{
			iqmmesh_t *mesh = &iqm->mesh[i];

			qglBufferSubData (GL_ELEMENT_ARRAY_BUFFER,
				iqm->indexoffsets[i],
				3 * mesh->num_triangles * sizeof (unsigned short),
				&iqm->tris[mesh->first_triangle]);
		}
	}

	qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, 0);
	qglBindBuffer (GL_ARRAY_BUFFER, 0);
}


