/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_world.c: world model rendering

// mh 2010-07-16 - replaced with RMQ engine version
// everything completely rewritten...
// this was horrible inefficient stuff; truly awful things lived in here
#include "quakedef.h"

#define SURFGETLIGHTING1(lm) \
	"TEX "lm", fragment.texcoord[1], texture[1], 2D; # red\n" \
	"MAD "lm", "lm", fragment.texcoord[2].x, fragment.texcoord[3];\n"

#define SURFGETLIGHTING3(lm) \
	"TEMP texr, texg, texb;\n" \
	"TEX texr, fragment.texcoord[1], texture[1], 2D; # red\n" \
	"TEX texg, fragment.texcoord[1], texture[2], 2D; # green\n" \
	"TEX texb, fragment.texcoord[1], texture[3], 2D; # blue\n" \
	"DP4 "lm".r, texr, fragment.texcoord[2]; # apply styles to red\n" \
	"DP4 "lm".g, texg, fragment.texcoord[2]; # apply styles to green\n" \
	"DP4 "lm".b, texb, fragment.texcoord[2]; # apply styles to blue\n" \
	"ADD "lm", "lm", fragment.texcoord[3]; # add dynamic light\n"

#define world_vp \
	"!!ARBvp1.0\n" \
	TRANSFORMVERTEX ("result.position", "vertex.attrib[0]") \
	"MOV result.texcoord[0], vertex.attrib[1]; # diffuse texcoords\n" \
	"MOV result.texcoord[1], vertex.attrib[2]; # lightmap texcoords\n" \
	"MOV result.texcoord[2], vertex.attrib[3]; # lightstyles\n" \
	"MOV result.texcoord[3], vertex.attrib[4]; # dynamic light addition\n" \
	"END\n"

#define WORLD_BUILDFP(getlight) \
	"!!ARBfp1.0\n" \
	"TEMP tex1, tex2;\n" \
	"TEX tex1, fragment.texcoord[0], texture[0], 2D;\n" \
	getlight \
	"MUL result.color, tex1, tex2;\n" \
	"END\n"

#define WORLD_BUILDFP_FB(getlight) \
	"!!ARBfp1.0\n" \
	"TEMP tex1, tex2, tex3;\n" \
	"TEX tex1, fragment.texcoord[0], texture[0], 2D;\n" \
	getlight \
	"TEX tex3, fragment.texcoord[0], texture[4], 2D;\n" \
	"MUL tex1, tex1, tex2;\n" \
	"MAX result.color, tex1, tex3;\n" \
	"END\n"

#define WORLD_BUILDRLIGHTMAP_FP(getlight) \
	"!!ARBfp1.0\n" \
	"TEMP tex1;\n" \
	getlight \
	"MUL result.color, tex1, 0.5;\n" \
	"END\n"

#define EXTRA_BUILDFP(getlight) \
	"!!ARBfp1.0\n" \
	"TEMP tex1, tex2;\n" \
	"TEX tex1, fragment.texcoord[0], texture[0], 2D;\n" \
	getlight \
	"MOV tex2.a, 1.0;\n" \
	"MUL tex2, tex2, program.env[11];\n" \
	"MUL result.color, tex1, tex2;\n" \
	"END\n"

#define EXTRA_BUILDFP_FB(getlight) \
	"!!ARBfp1.0\n" \
	"TEMP tex1, tex2, tex3;\n" \
	"TEX tex1, fragment.texcoord[0], texture[0], 2D;\n" \
	getlight \
	"MOV tex2.a, 1.0;\n" \
	"MUL tex2, tex2, program.env[11];\n" \
	"TEX tex3, fragment.texcoord[0], texture[4], 2D;\n" \
	"MUL tex3, tex3, program.env[11];\n" \
	"MUL tex1, tex1, tex2;\n" \
	"MAX result.color, tex1, tex3;\n" \
	"END\n"

#define EXTRA_BUILDRLIGHTMAP_FP(getlight) \
	"!!ARBfp1.0\n" \
	"TEMP tex1;\n" \
	getlight \
	"MOV tex1.a, 1.0;\n" \
	"MUL result.color, tex1, program.env[11];\n" \
	"END\n"

#define world_rfullbright_fp \
	"!!ARBfp1.0\n" \
	"TEX result.color, fragment.texcoord[0], texture[0], 2D;\n" \
	"END\n"

#define extra_rfullbright_fp \
	"!!ARBfp1.0\n" \
	"TEMP tex1;\n" \
	"TEX tex1, fragment.texcoord[0], texture[0], 2D;\n" \
	"MUL result.color, tex1, program.env[11];\n" \
	"END\n"

#define world_fp1 WORLD_BUILDFP (SURFGETLIGHTING1 ("tex2"))
#define world_fp3 WORLD_BUILDFP (SURFGETLIGHTING3 ("tex2"))

#define worldfb_fp1 WORLD_BUILDFP_FB (SURFGETLIGHTING1 ("tex2"))
#define worldfb_fp3 WORLD_BUILDFP_FB (SURFGETLIGHTING3 ("tex2"))

#define world_rlightmap_fp1 WORLD_BUILDRLIGHTMAP_FP (SURFGETLIGHTING1 ("tex1"))
#define world_rlightmap_fp3 WORLD_BUILDRLIGHTMAP_FP (SURFGETLIGHTING3 ("tex1"))

#define extra_fp1 EXTRA_BUILDFP (SURFGETLIGHTING1 ("tex2"))
#define extra_fp3 EXTRA_BUILDFP (SURFGETLIGHTING3 ("tex2"))

#define extrafb_fp1 EXTRA_BUILDFP_FB (SURFGETLIGHTING1 ("tex2"))
#define extrafb_fp3 EXTRA_BUILDFP_FB (SURFGETLIGHTING3 ("tex2"))

#define extra_rlightmap_fp1 EXTRA_BUILDRLIGHTMAP_FP (SURFGETLIGHTING1 ("tex1"))
#define extra_rlightmap_fp3 EXTRA_BUILDRLIGHTMAP_FP (SURFGETLIGHTING3 ("tex1"))

GLuint arb_world_vp = 0;

GLuint arb_world_fp1 = 0;
GLuint arb_worldfb_fp1 = 0;
GLuint arb_world_fp_fog1 = 0;
GLuint arb_worldfb_fp_fog1 = 0;

GLuint arb_world_fp3 = 0;
GLuint arb_worldfb_fp3 = 0;
GLuint arb_world_fp_fog3 = 0;
GLuint arb_worldfb_fp_fog3 = 0;

GLuint arb_world_rfullbright_fp = 0;
GLuint arb_world_rlightmap_fp1 = 0;
GLuint arb_world_rlightmap_fp3 = 0;

GLuint arb_extra_fp1 = 0;
GLuint arb_extrafb_fp1 = 0;
GLuint arb_extra_fp_fog1 = 0;
GLuint arb_extrafb_fp_fog1 = 0;

GLuint arb_extra_fp3 = 0;
GLuint arb_extrafb_fp3 = 0;
GLuint arb_extra_fp_fog3 = 0;
GLuint arb_extrafb_fp_fog3 = 0;

GLuint arb_extra_rfullbright_fp = 0;
GLuint arb_extra_rlightmap_fp1 = 0;
GLuint arb_extra_rlightmap_fp3 = 0;

void World_CreateShaders (void)
{
	arb_world_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, world_vp, false);

	arb_world_rfullbright_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, world_rfullbright_fp, true);
	arb_world_rlightmap_fp1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, world_rlightmap_fp1, true);
	arb_world_rlightmap_fp3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, world_rlightmap_fp3, true);

	arb_world_fp1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, world_fp1, false);
	arb_worldfb_fp1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, worldfb_fp1, false);
	arb_world_fp_fog1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, world_fp1, true);
	arb_worldfb_fp_fog1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, worldfb_fp1, true);

	arb_world_fp3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, world_fp3, false);
	arb_worldfb_fp3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, worldfb_fp3, false);
	arb_world_fp_fog3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, world_fp3, true);
	arb_worldfb_fp_fog3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, worldfb_fp3, true);

	arb_extra_fp1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extra_fp1, false);
	arb_extrafb_fp1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extrafb_fp1, false);
	arb_extra_fp_fog1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extra_fp1, true);
	arb_extrafb_fp_fog1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extrafb_fp1, true);

	arb_extra_fp3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extra_fp3, false);
	arb_extrafb_fp3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extrafb_fp3, false);
	arb_extra_fp_fog3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extra_fp3, true);
	arb_extrafb_fp_fog3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extrafb_fp3, true);

	arb_extra_rfullbright_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extra_rfullbright_fp, true);
	arb_extra_rlightmap_fp1 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extra_rlightmap_fp1, true);
	arb_extra_rlightmap_fp3 = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, extra_rlightmap_fp3, true);
}


void RWorld_SelectSurfacePrograms (msurface_t *surf, qboolean fullbright, qboolean extra, qboolean fog)
{
	surf->vertexprogram = arb_world_vp;

	if (surf->numstyles > 1)
	{
		if (!cl.worldmodel->lightdata || r_fullbright.value)
			surf->fragmentprogram = extra ? arb_extra_rfullbright_fp : arb_world_rfullbright_fp;
		else if (r_lightmap.value)
			surf->fragmentprogram = extra ? arb_extra_rlightmap_fp3 : arb_world_rlightmap_fp3;
		else if (fullbright && fog)
			surf->fragmentprogram = extra ? arb_extrafb_fp_fog3 : arb_worldfb_fp_fog3;
		else if (fullbright)
			surf->fragmentprogram = extra ? arb_extrafb_fp3 : arb_worldfb_fp3;
		else if (fog)
			surf->fragmentprogram = extra ? arb_extra_fp_fog3 : arb_world_fp_fog3;
		else surf->fragmentprogram = extra ? arb_extra_fp3 : arb_world_fp3;
	}
	else
	{
		if (!cl.worldmodel->lightdata || r_fullbright.value)
			surf->fragmentprogram = extra ? arb_extra_rfullbright_fp : arb_world_rfullbright_fp;
		else if (r_lightmap.value)
			surf->fragmentprogram = extra ? arb_extra_rlightmap_fp1 : arb_world_rlightmap_fp1;
		else if (fullbright && fog)
			surf->fragmentprogram = extra ? arb_extrafb_fp_fog1 : arb_worldfb_fp_fog1;
		else if (fullbright)
			surf->fragmentprogram = extra ? arb_extrafb_fp1 : arb_worldfb_fp1;
		else if (fog)
			surf->fragmentprogram = extra ? arb_extra_fp_fog1 : arb_world_fp_fog1;
		else surf->fragmentprogram = extra ? arb_extra_fp1 : arb_world_fp1;
	}
}


r_surfextra_t *r_surfextra = NULL;

GLuint r_surfacevbo = 0;

extern gltexture_t lightmap_textures[3][MAX_LIGHTMAPS];

extern cvar_t gl_fullbrights, r_drawflat, r_oldskyleaf, r_showtris; // johnfitz

byte *SV_FatPVS (vec3_t org, model_t *worldmodel);
extern byte mod_novis[MAX_MAP_LEAFS/8];
int vis_changed; // if true, force pvs to be refreshed

void R_DrawSkySurfaces (msurface_t *chain, entity_t *ent);
void R_DrawWaterSurfaces (msurface_t *chain, entity_t *ent, model_t *mod);
void Surf_SetAlpha (msurface_t *surf, entity_t *ent);


cvar_t r_normalmapping = {"r_normalmapping", "1"};
int r_normalmapregistration = false;

// mh 2010-07-16
// all surfaces from all models are collected into a modelsurfs list which may then be sorted, etc as we see fit
// this allows models to be batched together as efficiently as possible

entity_t r_worldentity;
extern mplane_t frustum[];

msurface_t **r_cachedsurfaces;
int r_numcachedsurfaces;

mleaf_t **r_cachedleafs;
int r_numcachedleafs;

qboolean r_rebuildworld = true;


// mh 2010-07-16 - this is called on a new map load (see r_newmap) to init the lists
// as the hunk will have been cleared from the previous map
void R_InitWorldCaches (void)
{
	SAFE_FREE (r_cachedsurfaces);
	SAFE_FREE (r_cachedleafs);

	r_numcachedsurfaces = 0;
	r_numcachedleafs = 0;

	r_cachedsurfaces = (msurface_t **) MallocZ (sizeof (msurface_t *) * cl.worldmodel->numsurfaces);
	r_cachedleafs = (mleaf_t **) MallocZ (sizeof (mleaf_t *) * (cl.worldmodel->numleafs + 1));

	r_rebuildworld = true;
}


entity_t *r_lastentity = NULL;
model_t *r_lastmodel = NULL;
glmatrix *r_lastmatrix = NULL;
texture_t *lasttexture = NULL;
int lastlightmap = -1;

float last_surfalpha = -1;
qboolean r_restart_surface = false;


qboolean R_UseInstancedLight (entity_t *ent);


extern cvar_t r_lavaalpha;
extern cvar_t r_telealpha;
extern cvar_t r_slimealpha;
extern cvar_t r_lockalpha;


float R_PlaneDist (mplane_t *plane, entity_t *ent)
{
	// for axial planes it's quicker to just eval the dist and there's no need to cache;
	// for non-axial we check the cache and use if it current, otherwise calc it and cache it
	if (plane->type < 3)
		return ent->modelorg[plane->type] - plane->dist;
	else if (plane->cacheframe == r_framecount && plane->cacheent == ent)
	{
		// Con_Printf ("Cached planedist\n");
		return plane->cachedist;
	}
	else plane->cachedist = DotProduct (ent->modelorg, plane->normal) - plane->dist;

	// cache the result for this plane
	plane->cacheframe = r_framecount;
	plane->cacheent = ent;

	// and return what we got
	return plane->cachedist;
}


msurface_t *r_lightmapchains[MAX_LIGHTMAPS];
msurface_t *r_skychain = NULL;
msurface_t *r_waterchain = NULL;
msurface_t **r_batchedsurfaces = NULL;

int r_maxbatchedsurfaces = 0;
int r_numbatchedsurfaces = 0;
int r_maxsurfindexes = 0;
int r_streamoffset = -1;

entity_t *r_lastdrawent = NULL;
entity_t *r_currentent = NULL;
gltexture_t *r_currentdiffuse = NULL;
gltexture_t *r_currentlightmap[3] = {NULL, NULL, NULL};
gltexture_t *r_currentluma = NULL;

GLuint r_currentvertexprogram = 0xffffffff;
GLuint r_currentfragmentprogram = 0xffffffff;

int r_firstvertex = 0x7fffffff;
int r_lastvertex = 0;

// qbsp guarantees no more than 64 verts per surf
unsigned short r_surfindexes[(64 - 2) * 3];


void R_ClearTextureChains (model_t *mod)
{
	int i;
	texture_t *t;

	for (i = 0; i < mod->numtextures; i++)
	{
		if (!(t = mod->textures[i])) continue;

		t->texturechain = NULL;
	}

	for (i = 0; i < MAX_LIGHTMAPS; i++)
		r_lightmapchains[i] = NULL;

	r_skychain = NULL;
	r_waterchain = NULL;
}


void R_ChainSurface (msurface_t *surf, entity_t *ent)
{
	if (surf->flags & SURF_DRAWSKY)
	{
		surf->texturechain = r_skychain;
		r_skychain = surf;
	}
	else if (surf->flags & SURF_DRAWTURB)
	{
		surf->lightmapchain = r_waterchain;
		r_waterchain = surf;
	}
	else
	{
		texture_t *tex = R_TextureAnimation (surf->texinfo->texture, ent ? ent->frame : 0);

		surf->texturechain = tex->texturechain;
		tex->texturechain = surf;
	}
}


void R_ResetBatch (void)
{
	r_firstvertex = 0x7fffffff;
	r_lastvertex = 0;
	r_numbatchedsurfaces = 0;
}


dlight_t *r_surflights[MAX_DLIGHTS];
extern int r_dlightframecount;

#define R_SurfUpdateLight4(dst, src) {dst[0] = src[0]; dst[1] = src[1]; dst[2] = src[2]; dst[3] = src[3];}
#define R_SurfUpdateLight3(dst, src) {dst[0] = src[0]; dst[1] = src[1]; dst[2] = src[2];}
#define R_SurfUpdateDynamic(dst, src, dyn) {dst[0] = src[0] + dyn[0]; dst[1] = src[1] + dyn[1]; dst[2] = src[2] + dyn[2];}

#define R_SurfDoDynamicLight(mvert) \
	if (mvert->dlightframe != r_dlightframecount) \
	{ \
		int lnum; \
		float f; \
		dlight_t *light; \
		float *rgb = mvert->dlrgb; \
		float *position = mvert->position; \
		for (lnum = 0, rgb[0] = rgb[1] = rgb[2] = 0; lnum < r_numsurflights; lnum++) \
		{ \
			light = r_surflights[lnum]; \
			f = VectorDistance2 (position, light->transformed); \
			if (f < light->r_radius) \
			{ \
				f = light->r_radius2 / (f + LIGHTOFFSET); \
				rgb[0] += light->r_rgb[0] * f; \
				rgb[1] += light->r_rgb[1] * f; \
				rgb[2] += light->r_rgb[2] * f; \
			} \
		} \
		mvert->dlightframe = r_dlightframecount; \
	}

void R_SetSurfaceStyles (msurface_t *surf, entity_t *ent)
{
	int i;
	r_surfextra_t *ex = surf->extra; //&r_surfextra[surf->streamoffset + surf->firstvertex];
	float styles[4] = {0.0f, 0.0f, 0.0f, 0.0f};
	float basedynamic[3] = {0.0f, 0.0f, 0.0f};
	float *dynamic;
	int r_numsurflights = 0;
	int r_cachedlight = 0;

	if (surf->flags & SURF_DRAWTURB) return;
	if (surf->flags & SURF_DRAWSKY) return;
	if (!cl.worldmodel->lightdata || r_fullbright.value) return;

	// get the dynamic lighting baseline
	if (!ent)
		dynamic = basedynamic;
	else dynamic = ent->aliasstate.shadelight;

	if (r_dynamic.value)
	{
		int maps, lnum;
		dlight_t *light;

		for (maps = 0; maps < MAX_SURFACE_STYLES && surf->styles[maps] != 255; maps++)
		{
			styles[maps] = d_lightstylevalue[surf->styles[maps]];

			if (surf->cached_light[maps] != styles[maps])
			{
				surf->cached_light[maps] = styles[maps];
				r_cachedlight |= (1 << maps);
			}
		}

		if (surf->dlightframe == r_dlightframecount)
		{
			// store all lights marked for this surf
			for (lnum = 0, r_numsurflights = 0, light = cl_dlights; lnum < MAX_DLIGHTS; lnum++, light++)
			{
				if (light->die < cl.time || light->radius < 0) continue;
				if (!(surf->dlightbits[lnum >> 5] & (1 << (lnum & 31)))) continue;

				r_surflights[r_numsurflights] = light;
				r_numsurflights++;
			}
		}
	}
	else
	{
		int maps;

		for (maps = 0; maps < MAX_SURFACE_STYLES && surf->styles[maps] != 255; maps++)
		{
			styles[maps] = 2.0f;

			if (surf->cached_light[maps] != styles[maps])
			{
				surf->cached_light[maps] = styles[maps];
				r_cachedlight |= (1 << maps);
			}
		}
	}

	// fixme - what do we not need to clear/update here...?
	if (r_numsurflights)
	{
		if (r_cachedlight)
		{
			// update styles and dynamic
			for (i = 0; i < surf->numvertexes; i++, ex++)
			{
				// pull from the previous or current calculation
				R_SurfUpdateLight4 (ex->styles, styles);
				R_SurfDoDynamicLight (surf->baseverts[i]);
				R_SurfUpdateDynamic (ex->dynamic, dynamic, surf->baseverts[i]->dlrgb);
			}
		}
		else
		{
			// update dynamic only
			for (i = 0; i < surf->numvertexes; i++, ex++)
			{
				// pull from the previous or current calculation
				R_SurfDoDynamicLight (surf->baseverts[i]);
				R_SurfUpdateDynamic (ex->dynamic, dynamic, surf->baseverts[i]->dlrgb);
			}
		}

		surf->cached_dlight = true;
	}
	else if (surf->cached_dlight)
	{
		if (r_cachedlight)
		{
			// update styles, clear dynamic
			for (i = 0; i < surf->numvertexes; i++, ex++)
			{
				R_SurfUpdateLight4 (ex->styles, styles);
				R_SurfUpdateLight3 (ex->dynamic, dynamic);
				surf->baseverts[i]->dlightframe = -1;
			}
		}
		else
		{
			// clear dynamic only
			for (i = 0; i < surf->numvertexes; i++, ex++)
			{
				R_SurfUpdateLight3 (ex->dynamic, dynamic);
				surf->baseverts[i]->dlightframe = -1;
			}
		}

		surf->cached_dlight = false;
	}
	else if (r_cachedlight)
	{
		// update styles only
		for (i = 0; i < surf->numvertexes; i++, ex++)
		{
			R_SurfUpdateLight4 (ex->styles, styles);
		}
	}
}


// we can't just memcpy the indexes as we need to add an offset to each so instead we'll Duff the bastards
// (8 seems optimal in testing)
#define R_TransferIndexes(offset, count, source) \
	{ \
		int n = (count + 7) >> 3; \
		unsigned short *src = source; \
		 \
		switch (count % 8) \
		{ \
		case 0: do {*ndx++ = offset + *src++; \
		case 7: *ndx++ = offset + *src++; \
		case 6: *ndx++ = offset + *src++; \
		case 5: *ndx++ = offset + *src++; \
		case 4: *ndx++ = offset + *src++; \
		case 3: *ndx++ = offset + *src++; \
		case 2: *ndx++ = offset + *src++; \
		case 1: *ndx++ = offset + *src++; \
		} while (--n > 0); \
		} \
	}


void R_FlushSurfaces (void)
{
	if (r_numbatchedsurfaces < 1)
	{
		// nothing batched (this can happen on a flush when we come in here first
		R_ResetBatch ();
		return;
	}

	if (r_lastdrawent != r_currentent)
	{
		// the dlight push should move to here (it will contain the transform) so that absolutely everything is valid
		// throughout the renderer
		if (r_currentent && r_currentent->headnode)
		{
			// just push the dlights (nodes will already have been marked)
			R_PushDlights (r_currentent->headnode, r_currentent);
		}

		r_lastdrawent = r_currentent;
	}

	if (r_numbatchedsurfaces > 1)
	{
		// multiple surfaces
		int s;
		unsigned short *ndx = (unsigned short *) scratchbuf;
		msurface_t *surf;
		int numindexes = 0;

		for (s = 0; s < r_numbatchedsurfaces; s++)
		{
			surf = r_batchedsurfaces[s];

			R_SetSurfaceStyles (surf, r_currentent);
			R_TransferIndexes (surf->firstvertex, surf->numindexes, r_surfindexes);
			numindexes += surf->numindexes;
		}

		GL_DrawElements (0, GL_TRIANGLES, r_firstvertex, r_lastvertex - 1, numindexes, GL_UNSIGNED_SHORT, scratchbuf);
	}
	else
	{
		// unbatched (can make this variable)
		int s;
		msurface_t *surf;

		for (s = 0; s < r_numbatchedsurfaces; s++)
		{
			surf = r_batchedsurfaces[s];

			R_SetSurfaceStyles (surf, r_currentent);
			GL_DrawArrays (GL_TRIANGLE_FAN, surf->firstvertex, surf->numvertexes);
		}
	}

	R_ResetBatch ();
}


void R_BeginSurfaces (void)
{
	R_ResetBatch ();

	r_streamoffset = -1;

	r_currentent = NULL;

	r_currentvertexprogram = 0xffffffff;
	r_currentfragmentprogram = 0xffffffff;

	// this can't be an entity or NULL as both are valid to test for changing to or from, so set it to something else
	r_lastdrawent = (entity_t *) r_batchedsurfaces;

	r_currentdiffuse = NULL;
	r_currentlightmap[0] = NULL;
	r_currentlightmap[1] = NULL;
	r_currentlightmap[2] = NULL;
	r_currentluma = NULL;
}


void R_EndSurfaces (void)
{
	R_FlushSurfaces ();

	if (r_currentent)
	{
		glLoadMatrixf (r_world_matrix.m16);
		r_currentent = NULL;
	}
}


void R_BatchSurface (msurface_t *surf)
{
	if (!surf->vertexprogram) return;
	if (!surf->fragmentprogram) return;

	if (surf->vertexprogram != r_currentvertexprogram)
	{
		R_FlushSurfaces ();
		GL_SetVertexProgram (surf->vertexprogram);
		r_currentvertexprogram = surf->vertexprogram;
	}

	if (surf->fragmentprogram != r_currentfragmentprogram)
	{
		R_FlushSurfaces ();
		GL_SetFragmentProgram (surf->fragmentprogram);
		r_currentfragmentprogram = surf->fragmentprogram;
	}

	// every surface now goes through this so it's a valid place to update the counter
	rs_brushpolys++;

	// batching
	if (surf->firstvertex < r_firstvertex) r_firstvertex = surf->firstvertex;
	if (surf->vertexrange > r_lastvertex) r_lastvertex = surf->vertexrange;

	if (r_numbatchedsurfaces < r_maxbatchedsurfaces)
	{
		r_batchedsurfaces[r_numbatchedsurfaces] = surf;
		r_numbatchedsurfaces++;
	}
}


#define STREAM_OFFSET(s) (void *) ((s * sizeof (float)) + (surf->streamoffset * sizeof (glvertex_t)))

void R_PrecheckSurface (msurface_t *surf, entity_t *ent)
{
	if (surf->streamoffset != r_streamoffset)
	{
		R_FlushSurfaces ();

		// select the correct vertex setup
		// OpenGL sucks here because it doesn't decouple layout from data specification
		if (surf->flags & SURF_DRAWSKY)
		{
			GL_EnableVertexArrays (VAA_0 | VAA_1);
			GL_VertexArrayPointer (r_surfacevbo, 0, 3, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (0));
			GL_VertexArrayPointer (r_surfacevbo, 1, 3, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (0));
		}
		else if (surf->flags & SURF_DRAWTURB)
		{
			GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2);
			GL_VertexArrayPointer (r_surfacevbo, 0, 3, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (0));
			GL_VertexArrayPointer (r_surfacevbo, 1, 2, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (3));
			GL_VertexArrayPointer (r_surfacevbo, 2, 2, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (5));
		}
		else if (!cl.worldmodel->lightdata || r_fullbright.value)
		{
			GL_EnableVertexArrays (VAA_0 | VAA_1);
			GL_VertexArrayPointer (r_surfacevbo, 0, 3, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (0));
			GL_VertexArrayPointer (r_surfacevbo, 1, 2, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (3));
		}
		else
		{
			GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2 | VAA_3 | VAA_4);

			GL_VertexArrayPointer (r_surfacevbo, 0, 3, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (0));
			GL_VertexArrayPointer (r_surfacevbo, 1, 2, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (3));
			GL_VertexArrayPointer (r_surfacevbo, 2, 2, GL_FLOAT, GL_FALSE, sizeof (glvertex_t), STREAM_OFFSET (5));

			GL_VertexArrayPointer (0, 3, 4, GL_FLOAT, GL_FALSE, sizeof (r_surfextra_t), r_surfextra[surf->streamoffset].styles);
			GL_VertexArrayPointer (0, 4, 3, GL_FLOAT, GL_FALSE, sizeof (r_surfextra_t), r_surfextra[surf->streamoffset].dynamic);
		}

		r_streamoffset = surf->streamoffset;

		// force surfs to recache their data (fixme - this will happen every frame, even when it shouldn't)
		// r_streamframe++;
	}

	if (ent != r_currentent)
	{
		R_FlushSurfaces ();

		if (ent)
			glLoadMatrixf (ent->matrix.m16);
		else glLoadMatrixf (r_world_matrix.m16);

		r_currentent = ent;
	}
}


void R_EmitSurface (msurface_t *surf, texture_t *tex, entity_t *ent)
{
	int j;
	gltexture_t *r_activelightmap[3] = {NULL, NULL, NULL};

	R_PrecheckSurface (surf, ent);

	if (R_UseInstancedLight (ent))
	{
		// lighting of instanced bmodels
		r_activelightmap[0] = r_blacktexture;
		r_activelightmap[1] = r_blacktexture;
		r_activelightmap[2] = r_blacktexture;
	}
	else if (!cl.worldmodel->lightdata || r_fullbright.value)
	{
		// make unlit maps go faster; the shader won't actually use these textures, they're just to save on state changes
		r_activelightmap[0] = r_greytexture;
		r_activelightmap[1] = r_greytexture;
		r_activelightmap[2] = r_greytexture;
	}
	else
	{
		// standard lighting
		r_activelightmap[0] = &lightmap_textures[0][surf->lightmaptexturenum];
		r_activelightmap[1] = &lightmap_textures[1][surf->lightmaptexturenum];
		r_activelightmap[2] = &lightmap_textures[2][surf->lightmaptexturenum];
	}

	if (r_currentdiffuse != tex->gltexture)
	{
		R_FlushSurfaces ();
		GL_BindTexture (GL_TEXTURE0, tex->gltexture);
		r_currentdiffuse = tex->gltexture;
	}

	for (j = 0; j < 3; j++)
	{
		if (r_currentlightmap[j] != r_activelightmap[j])
		{
			R_FlushSurfaces ();
			GL_BindTexture (GL_TEXTURE1 + j, r_activelightmap[j]);
			r_currentlightmap[j] = r_activelightmap[j];
		}
	}

	if (r_currentluma != tex->fullbright)
	{
		if (tex->fullbright)
		{
			R_FlushSurfaces ();
			GL_BindTexture (GL_TEXTURE4, tex->fullbright);
		}

		r_currentluma = tex->fullbright;
	}

	R_BatchSurface (surf);
}


void R_DrawSolidSurfaces (model_t *mod, entity_t *ent, qboolean fullbright)
{
	int i, j;
	texture_t *tex;
	msurface_t *surf;

	R_BeginSurfaces ();

	for (i = 0; i < mod->numtextures; i++)
	{
		if (!(tex = mod->textures[i])) continue;
		if (!(surf = tex->texturechain)) continue;

		if (tex->fullbright && !fullbright) continue;
		if (!tex->fullbright && fullbright) continue;

		// reorder by lightmap for cache coherence
		for (; surf; surf = surf->texturechain)
		{
			surf->lightmapchain = r_lightmapchains[surf->lightmaptexturenum];
			r_lightmapchains[surf->lightmaptexturenum] = surf;
		}

		// and draw in lightmap order
		// this also inverts the draw order from b2f to f2b (reasonably enough)
		for (j = 0; j < MAX_LIGHTMAPS; j++)
		{
			if (!(surf = r_lightmapchains[j])) continue;

			// no entity used as this entity has already been transformed
			for (; surf; surf = surf->lightmapchain)
			{
				RWorld_SelectSurfacePrograms (surf, fullbright, false, Fog_GetDensity () > 0);
				R_EmitSurface (surf, tex, ent);
			}

			R_FlushSurfaces ();
			r_lightmapchains[j] = NULL;
		}

		tex->texturechain = NULL;
	}
}


void R_DrawTextureChains (model_t *mod, entity_t *ent)
{
	if (r_skychain)
	{
		R_BeginSurfaces ();
		R_DrawSkySurfaces (r_skychain, ent);
		r_skychain = NULL;
	}

	R_DrawSolidSurfaces (mod, ent, false);
	R_DrawSolidSurfaces (mod, ent, true);

	if (r_waterchain)
	{
		R_BeginSurfaces ();
		R_DrawWaterSurfaces (r_waterchain, ent, mod);
		r_waterchain = NULL;
	}
}


void R_RecursiveWorldNode (mnode_t *node, int clipflags)
{
	float dot;
	int c, side;

	// the compiler should optimize this automatically anyway
tail_recurse:;
	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;

	if (clipflags)
	{
		for (c = 0, side = 0; c < 4; c++)
		{
			// clipflags is a 4 bit mask, for each bit 0 = auto accept on this side, 1 = need to test on this side
			// BOPS returns 0 (intersect), 1 (inside), or 2 (outside).  if a node is inside on any side then all
			// of it's child nodes are also guaranteed to be inside on the same side
			if (!(clipflags & (1 << c))) continue;
			if ((side = BOX_ON_PLANE_SIDE (node->minmaxs, node->minmaxs + 3, &frustum[c])) == 2) return;
			if (side == 1) clipflags &= ~(1 << c);
		}
	}

	if (node->contents < 0)
	{
		// node is a leaf so add stuff for drawing
		mleaf_t *leaf = (mleaf_t *) node;
		msurface_t **mark = leaf->firstmarksurface;
		int c = leaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				(*mark)->clipflags = clipflags;
				mark++;
			} while (--c);
		}

		if (leaf->efrags)
		{
			r_cachedleafs[r_numcachedleafs] = leaf;
			r_numcachedleafs++;
			R_StoreEfrags (&leaf->efrags);
		}

		return;
	}

	// figure which side of the node we're on
	dot = R_PlaneDist (node->plane, &r_worldentity);
	side = (dot >= 0 ? 0 : 1);

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side], clipflags);

	if ((c = node->numsurfaces) != 0)
	{
		msurface_t *surf = cl.worldmodel->surfaces + node->firstsurface;

		if (dot < 0 - BACKFACE_EPSILON)
			side = SURF_PLANEBACK;
		else if (dot > BACKFACE_EPSILON)
			side = 0;

		for (; c; c--, surf++)
		{
			if (surf->visframe != r_framecount) continue;
			if ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)) continue;		// wrong side

			// this can be optimized by only checking for culling if the leaf or node containing the surf intersects the frustum
			if (clipflags && surf->clipflags)
				if (R_CullBox (surf->mins, surf->maxs)) continue;

			// add to the list; the world is untransformed and has no matrix
			R_ChainSurface (surf, NULL);
			Surf_SetAlpha (surf, NULL);

			r_cachedsurfaces[r_numcachedsurfaces] = surf;
			r_numcachedsurfaces++;
		}
	}

	// recurse down the back side (the compiler should optimize this tail recursion)
	node = node->children[!side];
	goto tail_recurse;
	//R_RecursiveWorldNode (node->children[!node->side], clipflags);
}


void R_ExtraSurfsBegin (void)
{
	lasttexture = NULL;
	lastlightmap = -1;
	last_surfalpha = -1;

	GL_SetVertexProgram (arb_world_vp);
	R_BeginSurfaces ();
}


void R_ExtraSurfsEnd (void)
{
	R_EndSurfaces ();

	// because fences may have disabled it
	glEnable (GL_BLEND);
}


void R_RenderFenceTexture (msurface_t *surf, entity_t *ent, float surfalpha)
{
	texture_t *tex = R_TextureAnimation (surf->texinfo->texture, ent->frame);

	if (tex != lasttexture) r_restart_surface = true;
	if (surf->lightmaptexturenum != lastlightmap) r_restart_surface = true;
	if (surfalpha != last_surfalpha) r_restart_surface = true;

	RWorld_SelectSurfacePrograms (surf, tex->fullbright != NULL, true, Fog_GetDensity () > 0);

	if (r_restart_surface)
	{
		R_FlushSurfaces ();

		qglProgramEnvParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 11, 1, 1, 1, surfalpha);

		lasttexture = tex;
		lastlightmap = surf->lightmaptexturenum;
		last_surfalpha = surfalpha;
		r_restart_surface = false;
	}

	R_EmitSurface (surf, tex, ent);
}


void R_RenderAlphaSurface (msurface_t *surf, entity_t *ent, float surfalpha)
{
	texture_t *tex = R_TextureAnimation (surf->texinfo->texture, ent->frame);

	if (tex != lasttexture) r_restart_surface = true;
	if (surf->lightmaptexturenum != lastlightmap) r_restart_surface = true;
	if (surfalpha != last_surfalpha) r_restart_surface = true;

	RWorld_SelectSurfacePrograms (surf, tex->fullbright != NULL, true, Fog_GetDensity () > 0);

	if (r_restart_surface)
	{
		R_FlushSurfaces ();

		qglProgramEnvParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 11, 1, 1, 1, surfalpha);

		lasttexture = tex;
		lastlightmap = surf->lightmaptexturenum;
		last_surfalpha = surfalpha;
		r_restart_surface = false;
	}

	R_EmitSurface (surf, tex, ent);
}


// ==============================================================================
//
// SETUP CHAINS
//
// ==============================================================================

/*
===============
R_MarkLeafs -- johnfitz -- mark surfaces based on PVS and rebuild texture chains

it was a nice idea but you really to need to sort f2b to avoid horrible overdraw in big complex scenes
(and to get trivial culling of huge chunks of geometry)
===============
*/
void R_MarkLeafs (void)
{
	byte		*vis;
	mnode_t		*node;
	msurface_t	**mark;
	int			i;
	qboolean	nearwaterportal;
	extern		cvar_t r_lockpvs;

	// if visibility doesn't need regenerating just get out
	if (r_lockpvs.value && r_viewleaf) return;
	if (r_oldviewleaf == r_viewleaf && !vis_changed) return;

	r_visframecount++;
	r_oldviewleaf = r_viewleaf;

	// check this leaf for water portals
	// TODO: loop through all water surfs and use distance to leaf cullbox
	nearwaterportal = false;

	for (i = 0, mark = r_viewleaf->firstmarksurface; i < r_viewleaf->nummarksurfaces; i++, mark++)
		if ((*mark)->flags & SURF_DRAWTURB)
			nearwaterportal = true;

	// choose vis data
	if (r_novis.value || r_viewleaf->contents == CONTENTS_SOLID || r_viewleaf->contents == CONTENTS_SKY)
		vis = &mod_novis[0];
	else if (nearwaterportal)
		vis = SV_FatPVS (r_origin, cl.worldmodel);
	else vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);

	// mh 2010-07-18 - revert to the old way because fitz doesn't sort it's surfaces from front to back at all!!!
	// this gave us horrible overdraw in big complex scenes which more than wiped out the CPU gain from not using the old way
	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->visframe == r_visframecount)
					break;

				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}


/*
=============
R_DrawWorld -- johnfitz -- rewritten
=============
*/
void R_DrawBrushModel (entity_t *e);
void R_MergeBrushEntities (void);

void R_DrawWorld (void)
{
	int i;

	if (!r_drawworld.value || !r_refdef.drawworld) return;

	/*r_refdef.drawworld is for csqc. in this case bsp models won't be drawn. oh well. not drawing world is only really useful for models on your hud or whatever*/

	VectorCopy (r_refdef.vieworg, r_worldentity.modelorg);

	R_ClearTextureChains (cl.worldmodel);

	if (r_rebuildworld)
	{
		r_numcachedsurfaces = 0;
		r_numcachedleafs = 0;

		// emit the world surfs into modelsurfs
		R_RecursiveWorldNode (cl.worldmodel->nodes, 15);

		r_rebuildworld = false;
	}
	else
	{
		// bring everything else up to date
		for (i = 0; i < r_numcachedleafs; i++)
			R_StoreEfrags (&r_cachedleafs[i]->efrags);

		for (i = 0; i < r_numcachedsurfaces; i++)
		{
			msurface_t *surf = r_cachedsurfaces[i];

			// add to the list; the world is untransformed and has no matrix
			surf->visframe = r_framecount;
			R_ChainSurface (surf, NULL);
			Surf_SetAlpha (surf, NULL);
		}
	}

	// push dlights for the world
	R_PushDlights (cl.worldmodel->nodes, NULL);

	// draw world here....
	R_BeginSurfaces ();
	R_DrawTextureChains (cl.worldmodel, NULL);

	// add brush models to the list
	for (i = 0; i < cl_numvisedicts; i++)
	{
		entity_t *ent = cl_visedicts[i];

		// merged
		if (!r_drawentities.value) continue;

		if (ent->model->type == mod_brush)
		{
			// mh 2010-07-16 - this just adds to the list
			R_DrawBrushModel (ent);
		}
	}

	R_EndSurfaces ();

	// if (rs_dynamiclightmaps) Con_Printf ("%i lightmaps\n", rs_dynamiclightmaps);
}


#define MAX_STREAM_VERTS 60000

void RBrush_LoadVertexBuffers (void)
{
	int i, j, k, v;
	model_t	*m;
	int totalverts = 0;
	int numsurfaces = 0;
	glvertex_t surfverts[64];	// qbsp guarantees no more than 64 verts per surf
	unsigned short *ndx;
	int streamoffset = 0, streamverts = 0;

	r_maxsurfindexes = 0;

	for (j = 1; j < MAX_MODELS; j++)
	{
		msurface_t *surf = NULL;

		if (!(m = cl.model_precache[j])) break;
		if (m->type != mod_brush) continue;
		if (m->name[0] == '*') continue;

		// mark all surfaces in this model as using this vbo
		for (i = 0, surf = m->surfaces; i < m->numsurfaces; i++, surf++)
		{
			surf->model = m;
			totalverts += surf->numvertexes;
			r_maxsurfindexes += surf->numindexes;
			numsurfaces++;
		}
	}

	SAFE_FREE (r_batchedsurfaces);
	SAFE_FREE (r_surfextra);

	// alloc nothing if there was nothing to alloc
	if (!totalverts) return;

	r_batchedsurfaces = (msurface_t **) malloc (numsurfaces * sizeof (msurface_t *));
	r_maxbatchedsurfaces = numsurfaces;
	r_numbatchedsurfaces = 0;

	// create a new vbo for this model
	r_surfacevbo = GL_GetBuffer ();
	// Con_Printf ("allocated %i for surface VBO\n", r_surfacevbo);
	qglBindBuffer (GL_ARRAY_BUFFER, r_surfacevbo);
	qglBufferData (GL_ARRAY_BUFFER, totalverts * sizeof (glvertex_t), NULL, GL_STATIC_DRAW);

	// alloc extra space for GPU lightmapping data
	r_surfextra = (r_surfextra_t *) MallocZ (totalverts * sizeof (r_surfextra_t));

	// start counting again so that we can fixup first vertex relative to this vbo
	totalverts = 0;

	for (j = 1; j < MAX_MODELS; j++)
	{
		texture_t *tex = NULL;
		msurface_t *surf = NULL;

		if (!(m = cl.model_precache[j])) break;
		if (m->type != mod_brush) continue;
		if (m->name[0] == '*') continue;

		R_ClearTextureChains (m);

		for (i = 0, surf = m->surfaces; i < m->numsurfaces; i++, surf++)
		{
			// chain by texture
			surf->texturechain = surf->texinfo->texture->texturechain;
			surf->texinfo->texture->texturechain = surf;

			// force caches to rebuild
			surf->cached_light[0] = -666;
			surf->cached_dlight = true;
		}

		for (i = 0; i < m->numtextures; i++)
		{
			if (!(tex = m->textures[i])) continue;
			if (!(surf = tex->texturechain)) continue;

			for (k = 0; k < MAX_LIGHTMAPS; k++)
				r_lightmapchains[k] = NULL;

			for (; surf; surf = surf->texturechain)
			{
				// this is a fixup for sky and turb surfs
				if (surf->lightmaptexturenum < 0) surf->lightmaptexturenum = 0;

				surf->lightmapchain = r_lightmapchains[surf->lightmaptexturenum];
				r_lightmapchains[surf->lightmaptexturenum] = surf;
			}

			for (k = 0; k < MAX_LIGHTMAPS; k++)
			{
				if (!(surf = r_lightmapchains[k])) continue;

				for (; surf; surf = surf->lightmapchain)
				{
					// detect sky
					if (!Q_strncasecmp (surf->texinfo->texture->name, "sky", 3))
						surf->flags |= (SURF_DRAWSKY | SURF_DRAWTILED);
					else if (surf->flags & SURF_DRAWTURB)
						;
					else if (surf->texinfo->flags & TEX_MISSING)
					{
						if (surf->samples)
							surf->flags |= SURF_NOTEXTURE;
						else surf->flags |= (SURF_NOTEXTURE | SURF_DRAWTILED);
					}

					surf->midpoint[0] = surf->midpoint[1] = surf->midpoint[2] = 0;

					for (v = 0; v < surf->numvertexes; v++)
					{
						float st[2];
						float *vec = surf->baseverts[v]->position;

						VectorAdd (surf->midpoint, vec, surf->midpoint);

						surfverts[v].v[0] = vec[0];
						surfverts[v].v[1] = vec[1];
						surfverts[v].v[2] = vec[2];

						st[0] = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
						st[1] = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];

						if (surf->flags & SURF_DRAWTURB)
						{
							surfverts[v].st1[0] = (st[0] - surf->texinfo->vecs[0][3]) / 64.0f;
							surfverts[v].st1[1] = (st[1] - surf->texinfo->vecs[1][3]) / 64.0f;

							surfverts[v].st2[1] = st[0] - surf->texinfo->vecs[0][3];
							surfverts[v].st2[0] = st[1] - surf->texinfo->vecs[1][3];
						}
						else if (!(surf->flags & SURF_DRAWSKY))
						{
							surfverts[v].st1[0] = st[0] / surf->texinfo->texture->width;
							surfverts[v].st1[1] = st[1] / surf->texinfo->texture->height;

							surfverts[v].st2[0] = ((st[0] - surf->texturemins[0]) + (surf->light_s * 16) + 8) / (float) (LM_BLOCK_WIDTH * 16);
							surfverts[v].st2[1] = ((st[1] - surf->texturemins[1]) + (surf->light_t * 16) + 8) / (float) (LM_BLOCK_HEIGHT * 16);
						}
					}

					VectorScale (surf->midpoint, 1.0f / surf->numvertexes, surf->midpoint);

					qglBufferSubData (GL_ARRAY_BUFFER,
						streamverts * sizeof (glvertex_t),
						surf->numvertexes * sizeof (glvertex_t),
						surfverts);

					// partition the vertex buffer so that we can reasonably index into it
					if (totalverts + surf->numvertexes > MAX_STREAM_VERTS)
					{
						streamoffset = streamverts;
						totalverts = 0;
					}

					// update this to the actual VBO position
					surf->streamoffset = streamoffset;
					surf->firstvertex = totalverts;
					surf->extra = &r_surfextra[surf->streamoffset + surf->firstvertex];
					surf->vertexrange = surf->firstvertex + surf->numvertexes;

					totalverts += surf->numvertexes;
					streamverts += surf->numvertexes;
				}

				// how did this ever NOT crash?
				r_lightmapchains[k] = NULL;
			}

			tex->texturechain = NULL;
		}
	}

	qglBindBuffer (GL_ARRAY_BUFFER, 0);

	// qbsp guarantees no more than 64 verts per surf
	for (i = 2, ndx = r_surfindexes; i < 64; i++, ndx += 3)
	{
		ndx[0] = 0;
		ndx[1] = i - 1;
		ndx[2] = i;
	}
}

